# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import AsyncMock

import pytest

from airflow.providers.amazon.aws.hooks.comprehend import ComprehendHook
from airflow.providers.amazon.aws.triggers.comprehend import (
    ComprehendCreateDocumentClassifierCompletedTrigger,
    ComprehendPiiEntitiesDetectionJobCompletedTrigger,
)
from airflow.triggers.base import TriggerEvent

from unit.amazon.aws.utils.test_waiter import assert_expected_waiter_type

BASE_TRIGGER_CLASSPATH = "airflow.providers.amazon.aws.triggers.comprehend."


class TestBaseComprehendTrigger:
    EXPECTED_WAITER_NAME: str | None = None
    JOB_ID: str | None = None

    def test_setup(self):
        # Ensure that all subclasses have an expected waiter name set.
        if self.__class__.__name__ != "TestBaseComprehendTrigger":
            assert isinstance(self.EXPECTED_WAITER_NAME, str)
            assert isinstance(self.JOB_ID, str)


class TestComprehendPiiEntitiesDetectionJobCompletedTrigger(TestBaseComprehendTrigger):
    EXPECTED_WAITER_NAME = "pii_entities_detection_job_complete"
    JOB_ID = "job_id"

    def test_serialization(self):
        """Assert that arguments and classpath are correctly serialized."""
        trigger = ComprehendPiiEntitiesDetectionJobCompletedTrigger(job_id=self.JOB_ID)
        classpath, kwargs = trigger.serialize()
        assert classpath == BASE_TRIGGER_CLASSPATH + "ComprehendPiiEntitiesDetectionJobCompletedTrigger"
        assert kwargs.get("job_id") == self.JOB_ID

    @pytest.mark.asyncio
    @mock.patch.object(ComprehendHook, "get_waiter")
    @mock.patch.object(ComprehendHook, "get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.__aenter__.return_value = mock.MagicMock()
        mock_get_waiter().wait = AsyncMock()
        trigger = ComprehendPiiEntitiesDetectionJobCompletedTrigger(job_id=self.JOB_ID)

        generator = trigger.run()
        response = await generator.asend(None)

        assert response == TriggerEvent({"status": "success", "job_id": self.JOB_ID})
        assert_expected_waiter_type(mock_get_waiter, self.EXPECTED_WAITER_NAME)
        mock_get_waiter().wait.assert_called_once()


class TestComprehendCreateDocumentClassifierCompletedTrigger:
    EXPECTED_WAITER_NAME = "create_document_classifier_complete"
    DOCUMENT_CLASSIFIER_ARN = (
        "arn:aws:comprehend:us-east-1:123456789012:document-classifier/insurance-classifier/version/v1"
    )

    def test_serialization(self):
        """Assert that arguments and classpath are correctly serialized."""
        trigger = ComprehendCreateDocumentClassifierCompletedTrigger(
            document_classifier_arn=self.DOCUMENT_CLASSIFIER_ARN
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == BASE_TRIGGER_CLASSPATH + "ComprehendCreateDocumentClassifierCompletedTrigger"
        assert kwargs.get("document_classifier_arn") == self.DOCUMENT_CLASSIFIER_ARN

    @pytest.mark.asyncio
    @mock.patch.object(ComprehendHook, "get_waiter")
    @mock.patch.object(ComprehendHook, "get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.__aenter__.return_value = mock.MagicMock()
        mock_get_waiter().wait = AsyncMock()
        trigger = ComprehendCreateDocumentClassifierCompletedTrigger(
            document_classifier_arn=self.DOCUMENT_CLASSIFIER_ARN
        )

        generator = trigger.run()
        response = await generator.asend(None)

        assert response == TriggerEvent(
            {"status": "success", "document_classifier_arn": self.DOCUMENT_CLASSIFIER_ARN}
        )
        assert_expected_waiter_type(mock_get_waiter, self.EXPECTED_WAITER_NAME)
        mock_get_waiter().wait.assert_called_once()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from moto import mock_aws

from airflow.providers.amazon.aws.secrets.secrets_manager import SecretsManagerBackend


class TestSecretsManagerBackend:
    @mock.patch("airflow.providers.amazon.aws.secrets.secrets_manager.SecretsManagerBackend.get_conn_value")
    def test_aws_secrets_manager_get_connection(self, mock_get_value):
        mock_get_value.return_value = "scheme://user:pass@host:100"
        conn = SecretsManagerBackend().get_connection("fake_conn")
        assert conn.host == "host"

    @mock_aws
    def test_get_conn_value_full_url_mode(self):
        secret_id = "airflow/connections/test_postgres"
        create_param = {
            "Name": secret_id,
            "SecretString": "postgresql://airflow:airflow@host:5432/airflow",
        }

        secrets_manager_backend = SecretsManagerBackend()
        secrets_manager_backend.client.create_secret(**create_param)

        returned_uri = secrets_manager_backend.get_conn_value(conn_id="test_postgres")
        assert returned_uri == "postgresql://airflow:airflow@host:5432/airflow"

    @mock_aws
    def test_get_conn_value_non_existent_key(self):
        """
        Test that if the key with connection ID is not present,
        SecretsManagerBackend.get_connection should return None
        """
        conn_id = "test_mysql"

        secret_id = "airflow/connections/test_postgres"
        create_param = {
            "Name": secret_id,
            "SecretString": "postgresql://airflow:airflow@host:5432/airflow",
        }

        secrets_manager_backend = SecretsManagerBackend()
        secrets_manager_backend.client.create_secret(**create_param)

        assert secrets_manager_backend.get_conn_value(conn_id=conn_id) is None
        assert secrets_manager_backend.get_connection(conn_id=conn_id) is None

    @mock_aws
    def test_get_variable(self):
        secret_id = "airflow/variables/hello"
        create_param = {"Name": secret_id, "SecretString": "world"}

        secrets_manager_backend = SecretsManagerBackend()
        secrets_manager_backend.client.create_secret(**create_param)

        returned_uri = secrets_manager_backend.get_variable("hello")
        assert returned_uri == "world"

    @mock_aws
    def test_get_variable_non_existent_key(self):
        """
        Test that if Variable key is not present,
        SystemsManagerParameterStoreBackend.get_variables should return None
        """
        secret_id = "airflow/variables/hello"
        create_param = {"Name": secret_id, "SecretString": "world"}

        secrets_manager_backend = SecretsManagerBackend()
        secrets_manager_backend.client.create_secret(**create_param)

        assert secrets_manager_backend.get_variable("test_mysql") is None

    @mock_aws
    def test_get_config_non_existent_key(self):
        """
        Test that if Config key is not present,
        SystemsManagerParameterStoreBackend.get_config should return None
        """
        secret_id = "airflow/config/hello"
        create_param = {"Name": secret_id, "SecretString": "world"}

        secrets_manager_backend = SecretsManagerBackend()
        secrets_manager_backend.client.create_secret(**create_param)

        assert secrets_manager_backend.get_config("test") is None

    @mock.patch("airflow.providers.amazon.aws.secrets.secrets_manager.SecretsManagerBackend._get_secret")
    def test_connection_prefix_none_value(self, mock_get_secret):
        """
        Test that if Connection ID is not present in AWS Secrets Manager,
        SecretsManagerBackend.get_conn_value should return None,
        SecretsManagerBackend._get_secret should not be called
        """
        kwargs = {"connections_prefix": None}

        secrets_manager_backend = SecretsManagerBackend(**kwargs)

        assert secrets_manager_backend.get_conn_value("test_mysql") is None
        mock_get_secret.assert_not_called()

    @mock.patch("airflow.providers.amazon.aws.secrets.secrets_manager.SecretsManagerBackend._get_secret")
    def test_variable_prefix_none_value(self, mock_get_secret):
        """
        Test that if Variable key is not present in AWS Secrets Manager,
        SecretsManagerBackend.get_variables should return None,
        SecretsManagerBackend._get_secret should not be called
        """
        kwargs = {"variables_prefix": None}

        secrets_manager_backend = SecretsManagerBackend(**kwargs)

        assert secrets_manager_backend.get_variable("hello") is None
        mock_get_secret.assert_not_called()

    @mock.patch("airflow.providers.amazon.aws.secrets.secrets_manager.SecretsManagerBackend._get_secret")
    def test_config_prefix_none_value(self, mock_get_secret):
        """
        Test that if Config key is not present in AWS Secrets Manager,
        SecretsManagerBackend.get_config should return None,
        SecretsManagerBackend._get_secret should not be called
        """
        kwargs = {"config_prefix": None}

        secrets_manager_backend = SecretsManagerBackend(**kwargs)

        assert secrets_manager_backend.get_config("config") is None
        mock_get_secret.assert_not_called()

    @mock.patch(
        "airflow.providers.amazon.aws.secrets.secrets_manager.SecretsManagerBackend.client",
        new_callable=mock.PropertyMock,
    )
    @pytest.mark.parametrize(
        "connection_id, connections_lookup_pattern, num_client_calls",
        [
            ("test", "test", 1),
            ("test", ".*", 1),
            ("test", "T.*", 1),
            ("test", "dummy-pattern", 0),
            ("test", None, 1),
        ],
    )
    def test_connection_lookup_pattern(
        self, mock_client, connection_id, connections_lookup_pattern, num_client_calls
    ):
        """
        Test that if Connection ID is looked up in AWS Secrets Manager
        """
        mock_client().get_secret_value.return_value = {"SecretString": None}
        kwargs = {"connections_lookup_pattern": connections_lookup_pattern}

        secrets_manager_backend = SecretsManagerBackend(**kwargs)
        secrets_manager_backend.get_conn_value(connection_id)
        assert mock_client().get_secret_value.call_count == num_client_calls

    @mock.patch(
        "airflow.providers.amazon.aws.secrets.secrets_manager.SecretsManagerBackend.client",
        new_callable=mock.PropertyMock,
    )
    @pytest.mark.parametrize(
        "variable_key, variables_lookup_pattern, num_client_calls",
        [
            ("test", "test", 1),
            ("test", ".*", 1),
            ("test", "T.*", 1),
            ("test", "dummy-pattern", 0),
            ("test", None, 1),
        ],
    )
    def test_variable_lookup_pattern(
        self, mock_client, variable_key, variables_lookup_pattern, num_client_calls
    ):
        """
        Test that if Variable key is looked up in AWS Secrets Manager
        """
        mock_client().get_secret_value.return_value = {"SecretString": None}
        kwargs = {"variables_lookup_pattern": variables_lookup_pattern}

        secrets_manager_backend = SecretsManagerBackend(**kwargs)
        secrets_manager_backend.get_variable(variable_key)
        assert mock_client().get_secret_value.call_count == num_client_calls

    @mock.patch(
        "airflow.providers.amazon.aws.secrets.secrets_manager.SecretsManagerBackend.client",
        new_callable=mock.PropertyMock,
    )
    @pytest.mark.parametrize(
        "config_key, config_lookup_pattern, num_client_calls",
        [
            ("test", "test", 1),
            ("test", ".*", 1),
            ("test", "T.*", 1),
            ("test", "dummy-pattern", 0),
            ("test", None, 1),
        ],
    )
    def test_config_lookup_pattern(self, mock_client, config_key, config_lookup_pattern, num_client_calls):
        """
        Test that if Variable key is looked up in AWS Secrets Manager
        """
        mock_client().get_secret_value.return_value = {"SecretString": None}
        kwargs = {"config_lookup_pattern": config_lookup_pattern}

        secrets_manager_backend = SecretsManagerBackend(**kwargs)
        secrets_manager_backend.get_config(config_key)
        assert mock_client().get_secret_value.call_count == num_client_calls

    @mock.patch("airflow.providers.amazon.aws.hooks.base_aws.SessionFactory")
    def test_passing_client_kwargs(self, mock_session_factory):
        secrets_manager_backend = SecretsManagerBackend(
            use_ssl=False, role_arn="arn:aws:iam::222222222222:role/awesome-role", region_name="eu-central-1"
        )

        # Mock SessionFactory, session and client
        mock_session_factory_instance = mock_session_factory.return_value
        mock_ssm_client = mock.MagicMock(return_value="mock-secretsmanager-client")
        mock_session = mock.MagicMock()
        mock_session.client = mock_ssm_client
        mock_create_session = mock.MagicMock(return_value=mock_session)
        mock_session_factory_instance.create_session = mock_create_session

        secrets_manager_backend.client
        assert mock_session_factory.call_count == 1
        mock_session_factory_call_kwargs = mock_session_factory.call_args.kwargs
        assert "conn" in mock_session_factory_call_kwargs
        conn_wrapper = mock_session_factory_call_kwargs["conn"]

        assert conn_wrapper.conn_id == "SecretsManagerBackend__connection"
        assert conn_wrapper.role_arn == "arn:aws:iam::222222222222:role/awesome-role"
        assert conn_wrapper.region_name == "eu-central-1"

        mock_ssm_client.assert_called_once_with(
            service_name="secretsmanager", region_name="eu-central-1", use_ssl=False
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import MagicMock

import pytest

from airflow.models.dag import DAG
from airflow.providers.amazon.aws.operators.cloud_formation import (
    CloudFormationCreateStackOperator,
    CloudFormationDeleteStackOperator,
)
from airflow.utils import timezone

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

DEFAULT_DATE = timezone.datetime(2019, 1, 1)
DEFAULT_ARGS = {"owner": "airflow", "start_date": DEFAULT_DATE}


@pytest.fixture
def mocked_hook_client():
    with mock.patch("airflow.providers.amazon.aws.hooks.cloud_formation.CloudFormationHook.conn") as m:
        yield m


class TestCloudFormationCreateStackOperator:
    def test_init(self):
        op = CloudFormationCreateStackOperator(
            task_id="cf_create_stack_init",
            stack_name="fake-stack",
            cloudformation_parameters={},
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="eu-west-1",
            verify=True,
            botocore_config={"read_timeout": 42},
        )
        assert op.hook.client_type == "cloudformation"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "eu-west-1"
        assert op.hook._verify is True
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = CloudFormationCreateStackOperator(
            task_id="cf_create_stack_init",
            stack_name="fake-stack",
            cloudformation_parameters={},
        )
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    def test_create_stack(self, mocked_hook_client):
        stack_name = "myStack"
        timeout = 15
        template_body = "My stack body"

        operator = CloudFormationCreateStackOperator(
            task_id="test_task",
            stack_name=stack_name,
            cloudformation_parameters={"TimeoutInMinutes": timeout, "TemplateBody": template_body},
            dag=DAG("test_dag_id", schedule=None, default_args=DEFAULT_ARGS),
        )

        operator.execute(MagicMock())

        mocked_hook_client.create_stack.assert_any_call(
            StackName=stack_name, TemplateBody=template_body, TimeoutInMinutes=timeout
        )

    def test_template_fields(self):
        op = CloudFormationCreateStackOperator(
            task_id="cf_create_stack_init",
            stack_name="fake-stack",
            cloudformation_parameters={},
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="eu-west-1",
            verify=True,
            botocore_config={"read_timeout": 42},
        )

        validate_template_fields(op)

    def test_overwritten_conn_passed_to_hook(self):
        OVERWRITTEN_CONN = "new-conn-id"
        op = CloudFormationCreateStackOperator(
            task_id="cf_create_stack_pass_conn",
            stack_name="fake-stack",
            cloudformation_parameters={},
            aws_conn_id=OVERWRITTEN_CONN,
        )
        assert op.hook.aws_conn_id == OVERWRITTEN_CONN

    def test_default_conn_passed_to_hook(self):
        DEFAULT_CONN = "aws_default"
        op = CloudFormationCreateStackOperator(
            task_id="cf_create_stack_pass_default_conn", stack_name="fake-stack", cloudformation_parameters={}
        )
        assert op.hook.aws_conn_id == DEFAULT_CONN


class TestCloudFormationDeleteStackOperator:
    def test_init(self):
        op = CloudFormationDeleteStackOperator(
            task_id="cf_delete_stack_init",
            stack_name="fake-stack",
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="us-east-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.hook.client_type == "cloudformation"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "us-east-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = CloudFormationDeleteStackOperator(task_id="cf_delete_stack_init", stack_name="fake-stack")
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    def test_delete_stack(self, mocked_hook_client):
        stack_name = "myStackToBeDeleted"

        operator = CloudFormationDeleteStackOperator(
            task_id="test_task",
            stack_name=stack_name,
            dag=DAG("test_dag_id", schedule=None, default_args=DEFAULT_ARGS),
        )

        operator.execute(MagicMock())

        mocked_hook_client.delete_stack.assert_any_call(StackName=stack_name)

    def test_template_fields(self):
        op = CloudFormationDeleteStackOperator(
            task_id="cf_delete_stack_init",
            stack_name="fake-stack",
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="us-east-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )

        validate_template_fields(op)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import sys

from airflow.providers.amazon.aws.triggers.emr import (
    EmrAddStepsTrigger,
    EmrContainerTrigger,
    EmrCreateJobFlowTrigger,
    EmrServerlessCancelJobsTrigger,
    EmrServerlessCreateApplicationTrigger,
    EmrServerlessDeleteApplicationTrigger,
    EmrServerlessStartApplicationTrigger,
    EmrServerlessStartJobTrigger,
    EmrServerlessStopApplicationTrigger,
    EmrStepSensorTrigger,
    EmrTerminateJobFlowTrigger,
)


class TestEmrAddStepsTrigger:
    def test_serialization(self):
        job_flow_id = "test_job_flow_id"
        step_ids = ["step1", "step2"]
        waiter_delay = 10
        waiter_max_attempts = 5

        trigger = EmrAddStepsTrigger(
            job_flow_id=job_flow_id,
            step_ids=step_ids,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.emr.EmrAddStepsTrigger"
        assert kwargs == {
            "job_flow_id": "test_job_flow_id",
            "step_ids": ["step1", "step2"],
            "waiter_delay": 10,
            "waiter_max_attempts": 5,
            "aws_conn_id": "aws_default",
        }


class TestEmrCreateJobFlowTrigger:
    def test_serialization(self):
        job_flow_id = "test_job_flow_id"
        waiter_delay = 30
        waiter_max_attempts = 60
        aws_conn_id = "aws_default"

        trigger = EmrCreateJobFlowTrigger(
            job_flow_id=job_flow_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.emr.EmrCreateJobFlowTrigger"
        assert kwargs == {
            "job_flow_id": "test_job_flow_id",
            "waiter_delay": 30,
            "waiter_max_attempts": 60,
            "aws_conn_id": "aws_default",
        }


class TestEmrTerminateJobFlowTrigger:
    def test_serialization(self):
        job_flow_id = "test_job_flow_id"
        waiter_delay = 30
        waiter_max_attempts = 60
        aws_conn_id = "aws_default"

        trigger = EmrTerminateJobFlowTrigger(
            job_flow_id=job_flow_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.emr.EmrTerminateJobFlowTrigger"
        assert kwargs == {
            "job_flow_id": "test_job_flow_id",
            "waiter_delay": 30,
            "waiter_max_attempts": 60,
            "aws_conn_id": "aws_default",
        }


class TestEmrContainerTrigger:
    def test_serialization(self):
        virtual_cluster_id = "test_virtual_cluster_id"
        job_id = "test_job_id"
        waiter_delay = 30
        waiter_max_attempts = 600
        aws_conn_id = "aws_default"

        trigger = EmrContainerTrigger(
            virtual_cluster_id=virtual_cluster_id,
            job_id=job_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.emr.EmrContainerTrigger"
        assert kwargs == {
            "virtual_cluster_id": "test_virtual_cluster_id",
            "job_id": "test_job_id",
            "waiter_delay": 30,
            "waiter_max_attempts": 600,
            "aws_conn_id": "aws_default",
        }

    def test_serialization_default_max_attempts(self):
        virtual_cluster_id = "test_virtual_cluster_id"
        job_id = "test_job_id"
        waiter_delay = 30
        aws_conn_id = "aws_default"

        trigger = EmrContainerTrigger(
            virtual_cluster_id=virtual_cluster_id,
            job_id=job_id,
            waiter_delay=waiter_delay,
            aws_conn_id=aws_conn_id,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.emr.EmrContainerTrigger"
        assert kwargs == {
            "virtual_cluster_id": "test_virtual_cluster_id",
            "job_id": "test_job_id",
            "waiter_delay": 30,
            "waiter_max_attempts": sys.maxsize,
            "aws_conn_id": "aws_default",
        }


class TestEmrStepSensorTrigger:
    def test_serialization(self):
        job_flow_id = "test_job_flow_id"
        step_id = "test_step_id"
        waiter_delay = 30
        waiter_max_attempts = 60
        aws_conn_id = "aws_default"

        trigger = EmrStepSensorTrigger(
            job_flow_id=job_flow_id,
            step_id=step_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.emr.EmrStepSensorTrigger"
        assert kwargs == {
            "job_flow_id": "test_job_flow_id",
            "step_id": "test_step_id",
            "waiter_delay": 30,
            "waiter_max_attempts": 60,
            "aws_conn_id": "aws_default",
        }


class TestEmrServerlessCreateApplicationTrigger:
    def test_serialization(self):
        application_id = "test_application_id"
        waiter_delay = 30
        waiter_max_attempts = 60
        aws_conn_id = "aws_default"

        trigger = EmrServerlessCreateApplicationTrigger(
            application_id=application_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.emr.EmrServerlessCreateApplicationTrigger"
        assert kwargs == {
            "application_id": "test_application_id",
            "waiter_delay": 30,
            "waiter_max_attempts": 60,
            "aws_conn_id": "aws_default",
        }


class TestEmrServerlessStartApplicationTrigger:
    def test_serialization(self):
        application_id = "test_application_id"
        waiter_delay = 30
        waiter_max_attempts = 60
        aws_conn_id = "aws_default"

        trigger = EmrServerlessStartApplicationTrigger(
            application_id=application_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.emr.EmrServerlessStartApplicationTrigger"
        assert kwargs == {
            "application_id": "test_application_id",
            "waiter_delay": 30,
            "waiter_max_attempts": 60,
            "aws_conn_id": "aws_default",
        }


class TestEmrServerlessStopApplicationTrigger:
    def test_serialization(self):
        application_id = "test_application_id"
        waiter_delay = 30
        waiter_max_attempts = 60
        aws_conn_id = "aws_default"

        trigger = EmrServerlessStopApplicationTrigger(
            application_id=application_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.emr.EmrServerlessStopApplicationTrigger"
        assert kwargs == {
            "application_id": "test_application_id",
            "waiter_delay": 30,
            "waiter_max_attempts": 60,
            "aws_conn_id": "aws_default",
        }


class TestEmrServerlessStartJobTrigger:
    def test_serialization(self):
        application_id = "test_application_id"
        waiter_delay = 30
        waiter_max_attempts = 60
        job_id = "job_id"
        aws_conn_id = "aws_default"

        trigger = EmrServerlessStartJobTrigger(
            application_id=application_id,
            waiter_delay=waiter_delay,
            job_id=job_id,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.emr.EmrServerlessStartJobTrigger"
        assert kwargs == {
            "application_id": "test_application_id",
            "waiter_delay": 30,
            "waiter_max_attempts": 60,
            "job_id": "job_id",
            "aws_conn_id": "aws_default",
        }


class TestEmrServerlessDeleteApplicationTrigger:
    def test_serialization(self):
        application_id = "test_application_id"
        waiter_delay = 30
        waiter_max_attempts = 60
        aws_conn_id = "aws_default"

        trigger = EmrServerlessDeleteApplicationTrigger(
            application_id=application_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.emr.EmrServerlessDeleteApplicationTrigger"
        assert kwargs == {
            "application_id": "test_application_id",
            "waiter_delay": 30,
            "waiter_max_attempts": 60,
            "aws_conn_id": "aws_default",
        }


class TestEmrServerlessCancelJobsTrigger:
    def test_serialization(self):
        application_id = "test_application_id"
        waiter_delay = 30
        waiter_max_attempts = 60
        aws_conn_id = "aws_default"

        trigger = EmrServerlessCancelJobsTrigger(
            application_id=application_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.emr.EmrServerlessCancelJobsTrigger"
        assert kwargs == {
            "application_id": "test_application_id",
            "waiter_delay": 30,
            "waiter_max_attempts": 60,
            "aws_conn_id": "aws_default",
        }



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains AWS Firehose hook."""

from __future__ import annotations

from collections.abc import Iterable

from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook


class FirehoseHook(AwsBaseHook):
    """
    Interact with Amazon Kinesis Firehose.

    Provide thick wrapper around :external+boto3:py:class:`boto3.client("firehose") <Firehose.Client>`.

    :param delivery_stream: Name of the delivery stream

    Additional arguments (such as ``aws_conn_id``) may be specified and
    are passed down to the underlying AwsBaseHook.

    .. seealso::
        - :class:`airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook`
    """

    def __init__(self, delivery_stream: str, *args, **kwargs) -> None:
        self.delivery_stream = delivery_stream
        kwargs["client_type"] = "firehose"
        super().__init__(*args, **kwargs)

    def put_records(self, records: Iterable):
        """
        Write batch records to Kinesis Firehose.

        .. seealso::
            - :external+boto3:py:meth:`Firehose.Client.put_record_batch`

        :param records: list of records
        """
        return self.get_conn().put_record_batch(DeliveryStreamName=self.delivery_stream, Records=records)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.amazon.aws.hooks.redshift_data import (
    ABORTED_STATE,
    FAILED_STATE,
    RedshiftDataQueryAbortedError,
    RedshiftDataQueryFailedError,
)
from airflow.providers.amazon.aws.triggers.redshift_data import RedshiftDataTrigger
from airflow.triggers.base import TriggerEvent

TEST_CONN_ID = "aws_default"
TEST_TASK_ID = "123"
POLL_INTERVAL = 4.0


class TestRedshiftDataTrigger:
    def test_redshift_data_trigger_serialization(self):
        """
        Asserts that the RedshiftDataTrigger correctly serializes its arguments
        and classpath.
        """
        trigger = RedshiftDataTrigger(
            statement_id=[],
            task_id=TEST_TASK_ID,
            aws_conn_id=TEST_CONN_ID,
            poll_interval=POLL_INTERVAL,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.redshift_data.RedshiftDataTrigger"
        assert kwargs == {
            "statement_id": [],
            "task_id": TEST_TASK_ID,
            "poll_interval": POLL_INTERVAL,
            "aws_conn_id": TEST_CONN_ID,
            "region_name": None,
            "botocore_config": None,
            "verify": None,
        }

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "return_value, response",
        [
            (
                True,
                TriggerEvent({"status": "success", "statement_id": "uuid"}),
            ),
            (
                False,
                TriggerEvent(
                    {"status": "error", "message": f"{TEST_TASK_ID} failed", "statement_id": "uuid"}
                ),
            ),
        ],
    )
    @mock.patch(
        "airflow.providers.amazon.aws.hooks.redshift_data.RedshiftDataHook.check_query_is_finished_async"
    )
    @mock.patch(
        "airflow.providers.amazon.aws.hooks.redshift_data.RedshiftDataHook.is_still_running",
        return_value=False,
    )
    async def test_redshift_data_trigger_run(
        self, mocked_is_still_running, mock_check_query_is_finised_async, return_value, response
    ):
        """
        Tests that RedshiftDataTrigger only fires once the query execution reaches a successful state.
        """
        mock_check_query_is_finised_async.return_value = return_value
        trigger = RedshiftDataTrigger(
            statement_id="uuid",
            task_id=TEST_TASK_ID,
            poll_interval=POLL_INTERVAL,
            aws_conn_id=TEST_CONN_ID,
        )
        generator = trigger.run()
        actual = await generator.asend(None)
        assert response == actual

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "raised_exception, expected_response",
        [
            (
                RedshiftDataQueryFailedError("Failed"),
                {
                    "status": "error",
                    "statement_id": "uuid",
                    "message": "Failed",
                    "type": FAILED_STATE,
                },
            ),
            (
                RedshiftDataQueryAbortedError("Aborted"),
                {
                    "status": "error",
                    "statement_id": "uuid",
                    "message": "Aborted",
                    "type": ABORTED_STATE,
                },
            ),
            (
                Exception(f"{TEST_TASK_ID} failed"),
                {"status": "error", "statement_id": "uuid", "message": f"{TEST_TASK_ID} failed"},
            ),
        ],
    )
    @mock.patch(
        "airflow.providers.amazon.aws.hooks.redshift_data.RedshiftDataHook.check_query_is_finished_async"
    )
    @mock.patch(
        "airflow.providers.amazon.aws.hooks.redshift_data.RedshiftDataHook.is_still_running",
        return_value=False,
    )
    async def test_redshift_data_trigger_exception(
        self, mocked_is_still_running, mock_check_query_is_finised_async, raised_exception, expected_response
    ):
        """
        Test that RedshiftDataTrigger fires the correct event in case of an error.
        """
        mock_check_query_is_finised_async.side_effect = raised_exception

        trigger = RedshiftDataTrigger(
            statement_id="uuid",
            task_id=TEST_TASK_ID,
            poll_interval=POLL_INTERVAL,
            aws_conn_id=TEST_CONN_ID,
        )
        task = [i async for i in trigger.run()]
        assert len(task) == 1
        assert TriggerEvent(expected_response) in task



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.mwaa import MwaaHook
from airflow.providers.amazon.aws.sensors.mwaa import MwaaDagRunSensor
from airflow.utils.state import DagRunState

SENSOR_KWARGS = {
    "task_id": "test_mwaa_sensor",
    "external_env_name": "test_env",
    "external_dag_id": "test_dag",
    "external_dag_run_id": "test_run_id",
    "deferrable": False,
    "poke_interval": 5,
    "max_retries": 100,
}

SENSOR_STATE_KWARGS = {
    "success_states": ["a", "b"],
    "failure_states": ["c", "d"],
}


@pytest.fixture
def mock_invoke_rest_api():
    with mock.patch.object(MwaaHook, "invoke_rest_api") as m:
        yield m


class TestMwaaDagRunSuccessSensor:
    def test_init_success(self):
        sensor = MwaaDagRunSensor(**SENSOR_KWARGS, **SENSOR_STATE_KWARGS)
        assert sensor.external_env_name == SENSOR_KWARGS["external_env_name"]
        assert sensor.external_dag_id == SENSOR_KWARGS["external_dag_id"]
        assert sensor.external_dag_run_id == SENSOR_KWARGS["external_dag_run_id"]
        assert set(sensor.success_states) == set(SENSOR_STATE_KWARGS["success_states"])
        assert set(sensor.failure_states) == set(SENSOR_STATE_KWARGS["failure_states"])
        assert sensor.deferrable == SENSOR_KWARGS["deferrable"]
        assert sensor.poke_interval == SENSOR_KWARGS["poke_interval"]
        assert sensor.max_retries == SENSOR_KWARGS["max_retries"]

        sensor = MwaaDagRunSensor(**SENSOR_KWARGS)
        assert sensor.success_states == {DagRunState.SUCCESS.value}
        assert sensor.failure_states == {DagRunState.FAILED.value}

    def test_init_failure(self):
        with pytest.raises(ValueError, match=r".*success_states.*failure_states.*"):
            MwaaDagRunSensor(
                **SENSOR_KWARGS, success_states={"state1", "state2"}, failure_states={"state2", "state3"}
            )

    @pytest.mark.parametrize("state", SENSOR_STATE_KWARGS["success_states"])
    def test_poke_completed(self, mock_invoke_rest_api, state):
        mock_invoke_rest_api.return_value = {"RestApiResponse": {"state": state}}
        assert MwaaDagRunSensor(**SENSOR_KWARGS, **SENSOR_STATE_KWARGS).poke({})

    @pytest.mark.parametrize("state", ["e", "f"])
    def test_poke_not_completed(self, mock_invoke_rest_api, state):
        mock_invoke_rest_api.return_value = {"RestApiResponse": {"state": state}}
        assert not MwaaDagRunSensor(**SENSOR_KWARGS, **SENSOR_STATE_KWARGS).poke({})

    @pytest.mark.parametrize("state", SENSOR_STATE_KWARGS["failure_states"])
    def test_poke_terminated(self, mock_invoke_rest_api, state):
        mock_invoke_rest_api.return_value = {"RestApiResponse": {"state": state}}
        with pytest.raises(AirflowException, match=f".*{state}.*"):
            MwaaDagRunSensor(**SENSOR_KWARGS, **SENSOR_STATE_KWARGS).poke({})



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""This module contains a web hook for Chime."""

from __future__ import annotations

import json
import re
from functools import cached_property
from typing import Any

from airflow.exceptions import AirflowException
from airflow.providers.http.hooks.http import HttpHook


class ChimeWebhookHook(HttpHook):
    """
    Interact with Amazon Chime Webhooks to create notifications.

    .. warning:: This hook is only designed to work with web hooks and not chat bots.

    :param chime_conn_id: :ref:`Amazon Chime Connection ID <howto/connection:chime>`
        with Endpoint as `https://hooks.chime.aws` and the webhook token
        in the form of ``{webhook.id}?token{webhook.token}``
    """

    conn_name_attr = "chime_conn_id"
    default_conn_name = "chime_default"
    conn_type = "chime"
    hook_name = "Amazon Chime Webhook"

    def __init__(
        self,
        chime_conn_id: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._chime_conn_id = chime_conn_id

    @cached_property
    def webhook_endpoint(self):
        return self._get_webhook_endpoint(self._chime_conn_id)

    def _get_webhook_endpoint(self, conn_id: str) -> str:
        """
        Given a Chime conn_id return the default webhook endpoint.

        :param conn_id: The provided connection ID.
        :return: Endpoint(str) for chime webhook.
        """
        conn = self.get_connection(conn_id)
        token = conn.password
        if token is None:
            raise AirflowException("Webhook token field is missing and is required.")
        if not conn.schema:
            raise AirflowException("Webook schema field is missing and is required")
        if not conn.host:
            raise AirflowException("Webhook host field is missing and is required.")
        url = conn.schema + "://" + conn.host
        endpoint = url + token
        # Check to make sure the endpoint matches what Chime expects
        if not re.fullmatch(r"[a-zA-Z0-9_-]+\?token=[a-zA-Z0-9_-]+", token):
            raise AirflowException(
                "Expected Chime webhook token in the form of '{webhook.id}?token={webhook.token}'."
            )

        return endpoint

    def _build_chime_payload(self, message: str) -> str:
        """
        Build payload for Chime and ensures messages do not exceed max length allowed.

        :param message: The message you want to send to your Chime room. (max 4096 characters)
        """
        payload: dict[str, Any] = {}
        # We need to make sure that the message does not exceed the max length for Chime
        if len(message) > 4096:
            raise AirflowException("Chime message must be 4096 characters or less.")

        payload["Content"] = message
        return json.dumps(payload)

    def send_message(self, message: str) -> None:
        """
        Execute calling the Chime webhook endpoint.

        :param message: The message you want to send to your Chime room.(max 4096 characters)
        """
        chime_payload = self._build_chime_payload(message)
        self.run(
            endpoint=self.webhook_endpoint, data=chime_payload, headers={"Content-type": "application/json"}
        )

    @classmethod
    def get_ui_field_behaviour(cls) -> dict[str, Any]:
        """Return custom field behaviour to only get what is needed for Chime webhooks to function."""
        return {
            "hidden_fields": ["login", "port", "extra"],
            "relabeling": {
                "host": "Chime Webhook Endpoint",
                "password": "Chime Webhook token",
            },
            "placeholders": {
                "schema": "https",
                "host": "hooks.chime.aws/incomingwebhook/",
                "password": "T00000000?token=XXXXXXXXXXXXXXXXXXXXXXXX",
            },
        }



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from functools import cached_property
from typing import TYPE_CHECKING, Any

import pyathena
from sqlalchemy.engine.url import URL

from airflow.exceptions import AirflowException, AirflowNotFoundException
from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook
from airflow.providers.amazon.aws.utils.connection_wrapper import AwsConnectionWrapper
from airflow.providers.common.sql.hooks.sql import DbApiHook

if TYPE_CHECKING:
    from pyathena.connection import Connection as AthenaConnection


class AthenaSQLHook(AwsBaseHook, DbApiHook):
    """
    Interact with Amazon Athena.

    Provide wrapper around PyAthena library.

    :param athena_conn_id: :ref:`Amazon Athena Connection <howto/connection:athena>`.

    Additional arguments (such as ``aws_conn_id``) may be specified and
    are passed down to the underlying AwsBaseHook.

    You can specify ``driver`` in ``extra`` of your connection in order to use
    a different driver than the default ``rest``.

    Also, aws_domain could be specified in ``extra`` of your connection.

    PyAthena and AWS Authentication parameters could be passed in extra field of ``athena_conn_id`` connection.

    Passing authentication parameters in ``athena_conn_id`` will override those in ``aws_conn_id``.

    .. seealso::
        :class:`~airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook`

    .. note::
        get_uri() depends on SQLAlchemy and PyAthena.
    """

    conn_name_attr = "athena_conn_id"
    default_conn_name = "athena_default"
    conn_type = "athena"
    hook_name = "Amazon Athena"
    supports_autocommit = True

    def __init__(self, athena_conn_id: str = default_conn_name, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.athena_conn_id = athena_conn_id

    @classmethod
    def get_ui_field_behaviour(cls) -> dict[str, Any]:
        """Return custom UI field behaviour for AWS Athena Connection."""
        return {
            "hidden_fields": ["host", "port"],
            "relabeling": {
                "login": "AWS Access Key ID",
                "password": "AWS Secret Access Key",
            },
            "placeholders": {
                "login": "AKIAIOSFODNN7EXAMPLE",
                "password": "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY",
                "extra": json.dumps(
                    {
                        "aws_domain": "amazonaws.com",
                        "driver": "rest",
                        "s3_staging_dir": "s3://bucket_name/staging/",
                        "work_group": "primary",
                        "region_name": "us-east-1",
                        "session_kwargs": {"profile_name": "default"},
                        "config_kwargs": {"retries": {"mode": "standard", "max_attempts": 10}},
                        "role_arn": "arn:aws:iam::123456789098:role/role-name",
                        "assume_role_method": "assume_role",
                        "assume_role_kwargs": {"RoleSessionName": "airflow"},
                        "aws_session_token": "AQoDYXdzEJr...EXAMPLETOKEN",
                        "endpoint_url": "http://localhost:4566",
                    },
                    indent=2,
                ),
            },
        }

    @cached_property
    def conn_config(self) -> AwsConnectionWrapper:
        """Get the Airflow Connection object and wrap it in helper (cached)."""
        athena_conn = self.get_connection(self.athena_conn_id)
        if self.aws_conn_id:
            try:
                connection = self.get_connection(self.aws_conn_id)
                connection.login = athena_conn.login
                connection.password = athena_conn.password
                connection.schema = athena_conn.schema
                merged_extra = {**athena_conn.extra_dejson, **connection.extra_dejson}
                try:
                    extra_json = json.dumps(merged_extra)
                    connection.extra = extra_json
                except (TypeError, ValueError):
                    raise ValueError(
                        f"Encountered non-JSON in `extra` field for connection {self.aws_conn_id!r}."
                    )
            except AirflowNotFoundException:
                connection = athena_conn
                connection.conn_type = "aws"
                self.log.warning(
                    "Unable to find AWS Connection ID '%s', switching to empty.", self.aws_conn_id
                )

        return AwsConnectionWrapper(
            conn=connection,
            region_name=self._region_name,
            botocore_config=self._config,
            verify=self._verify,
        )

    @property
    def conn(self) -> AwsConnectionWrapper:
        """Get Aws Connection Wrapper object."""
        return self.conn_config

    def _get_conn_params(self) -> dict[str, str | None]:
        """Retrieve connection parameters."""
        if not self.conn.region_name:
            raise AirflowException("region_name must be specified in the connection's extra")

        return dict(
            driver=self.conn.extra_dejson.get("driver", "rest"),
            schema_name=self.conn.schema,
            region_name=self.conn.region_name,
            aws_domain=self.conn.extra_dejson.get("aws_domain", "amazonaws.com"),
        )

    def get_uri(self) -> str:
        """Overridden to use the Athena dialect as driver name."""
        conn_params = self._get_conn_params()
        creds = self.get_credentials(region_name=conn_params["region_name"])

        return URL.create(
            f"awsathena+{conn_params['driver']}",
            username=creds.access_key,
            password=creds.secret_key,
            host=f"athena.{conn_params['region_name']}.{conn_params['aws_domain']}",
            port=443,
            database=conn_params["schema_name"],
            query={"aws_session_token": creds.token, **self.conn.extra_dejson},
        )

    def get_conn(self) -> AthenaConnection:
        """Get a ``pyathena.Connection`` object."""
        conn_params = self._get_conn_params()

        conn_kwargs: dict = {
            "schema_name": conn_params["schema_name"],
            "region_name": conn_params["region_name"],
            "session": self.get_session(region_name=conn_params["region_name"]),
            **self.conn.extra_dejson,
        }

        return pyathena.connect(**conn_kwargs)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Generator
from unittest import mock

import pytest
from boto3 import client
from moto import mock_aws

from airflow.exceptions import AirflowException, TaskDeferred
from airflow.providers.amazon.aws.hooks.neptune import NeptuneHook
from airflow.providers.amazon.aws.operators.neptune import (
    NeptuneStartDbClusterOperator,
    NeptuneStopDbClusterOperator,
)

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

CLUSTER_ID = "test_cluster"

EXPECTED_RESPONSE = {"db_cluster_id": CLUSTER_ID}


@pytest.fixture
def hook() -> Generator[NeptuneHook, None, None]:
    with mock_aws():
        yield NeptuneHook(aws_conn_id="aws_default")


@pytest.fixture
def _create_cluster(hook: NeptuneHook):
    hook.conn.create_db_cluster(
        DBClusterIdentifier=CLUSTER_ID,
        Engine="neptune",
    )
    if not hook.conn.describe_db_clusters()["DBClusters"]:
        raise ValueError("AWS not properly mocked")


class TestNeptuneStartClusterOperator:
    @mock.patch.object(NeptuneHook, "conn")
    @mock.patch.object(NeptuneHook, "get_waiter")
    def test_start_cluster_wait_for_completion(self, mock_hook_get_waiter, mock_conn):
        operator = NeptuneStartDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=False,
            wait_for_completion=True,
            aws_conn_id="aws_default",
        )

        resp = operator.execute(None)
        mock_hook_get_waiter.assert_called_once_with("cluster_available")
        assert resp == EXPECTED_RESPONSE

    @mock.patch.object(NeptuneHook, "conn")
    @mock.patch.object(NeptuneHook, "get_waiter")
    def test_start_cluster_no_wait(self, mock_hook_get_waiter, mock_conn):
        operator = NeptuneStartDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=False,
            wait_for_completion=False,
            aws_conn_id="aws_default",
        )

        resp = operator.execute(None)
        mock_hook_get_waiter.assert_not_called()
        assert resp == EXPECTED_RESPONSE

    @mock.patch.object(NeptuneHook, "conn")
    @mock.patch.object(NeptuneHook, "get_cluster_status")
    @mock.patch.object(NeptuneHook, "get_waiter")
    def test_start_cluster_cluster_available(self, mock_waiter, mock_get_cluster_status, mock_conn):
        mock_get_cluster_status.return_value = "available"
        operator = NeptuneStartDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=False,
            wait_for_completion=True,
            aws_conn_id="aws_default",
        )

        resp = operator.execute(None)

        mock_conn.start_db_cluster.assert_not_called()
        mock_waiter.assert_not_called()
        assert resp == {"db_cluster_id": CLUSTER_ID}

    @mock.patch.object(NeptuneHook, "conn")
    def test_start_cluster_deferrable(self, mock_conn):
        operator = NeptuneStartDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=True,
            wait_for_completion=False,
            aws_conn_id="aws_default",
        )

        with pytest.raises(TaskDeferred):
            operator.execute(None)

    @mock.patch.object(NeptuneHook, "conn")
    @mock.patch.object(NeptuneHook, "get_cluster_status")
    @mock.patch.object(NeptuneHook, "get_waiter")
    def test_start_cluster_cluster_error(self, mock_waiter, mock_get_cluster_status, mock_conn):
        mock_get_cluster_status.return_value = "migration-failed"
        operator = NeptuneStartDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=False,
            wait_for_completion=True,
            aws_conn_id="aws_default",
        )

        with pytest.raises(AirflowException):
            operator.execute(None)

    @mock.patch("airflow.providers.amazon.aws.operators.neptune.NeptuneStartDbClusterOperator.defer")
    @mock.patch("airflow.providers.amazon.aws.operators.neptune.handle_waitable_exception")
    @mock.patch.object(NeptuneHook, "conn")
    def test_start_cluster_not_ready_defer(self, mock_conn, mock_wait, mock_defer):
        err_response = {"Error": {"Code": "InvalidClusterState", "Message": "Test message"}}
        exception = client("neptune").exceptions.ClientError(err_response, "test")
        returned_exception = type(exception)

        mock_conn.exceptions.InvalidClusterStateFault = returned_exception
        mock_conn.start_db_cluster.side_effect = exception
        operator = NeptuneStartDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=True,
            wait_for_completion=False,
            aws_conn_id="aws_default",
        )

        operator.execute(None)

        mock_wait.assert_called_once_with(
            operator=operator,
            err="InvalidClusterState",
        )
        assert mock_defer.call_count == 1

    @mock.patch.object(NeptuneHook, "get_waiter")
    @mock.patch.object(NeptuneHook, "conn")
    def test_start_cluster_instances_not_ready(self, mock_conn, mock_get_waiter):
        err_response = {"Error": {"Code": "InvalidDBInstanceState", "Message": "Test message"}}
        exception = client("neptune").exceptions.ClientError(err_response, "test")
        returned_exception = type(exception)

        mock_conn.exceptions.InvalidClusterStateFault = returned_exception
        mock_conn.start_db_cluster.side_effect = exception

        operator = NeptuneStartDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=False,
            wait_for_completion=False,
            aws_conn_id="aws_default",
        )
        operator.execute(None)
        mock_get_waiter.assert_any_call("db_instance_available")

    @mock.patch("airflow.providers.amazon.aws.operators.neptune.NeptuneStartDbClusterOperator.defer")
    @mock.patch.object(NeptuneHook, "conn")
    def test_start_cluster_instances_not_ready_defer(self, mock_conn, mock_defer):
        """Tests both waiters are called if an instance exception is raised"""

        err_response = {"Error": {"Code": "InvalidDBInstanceState", "Message": "Test message"}}
        exception = client("neptune").exceptions.ClientError(err_response, "test")
        returned_exception = type(exception)

        mock_conn.exceptions.InvalidClusterStateFault = returned_exception
        mock_conn.start_db_cluster.side_effect = exception
        operator = NeptuneStartDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=True,
            wait_for_completion=False,
            aws_conn_id="aws_default",
        )

        operator.execute(None)

        # mock_defer.assert_has_calls(calls)
        assert mock_defer.call_count == 2

    def test_template_fields(self):
        operator = NeptuneStartDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=True,
            wait_for_completion=False,
            aws_conn_id="aws_default",
        )
        validate_template_fields(operator)


class TestNeptuneStopClusterOperator:
    @mock.patch.object(NeptuneHook, "conn")
    @mock.patch.object(NeptuneHook, "get_cluster_status")
    @mock.patch.object(NeptuneHook, "get_waiter")
    def test_stop_cluster_wait_for_completion(self, mock_hook_get_waiter, mock_get_cluster_status, mock_conn):
        '''Test the waiter is only once when the cluster is "available"'''
        mock_get_cluster_status.return_value = "available"
        operator = NeptuneStopDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=False,
            wait_for_completion=True,
            aws_conn_id="aws_default",
        )

        resp = operator.execute(None)
        mock_hook_get_waiter.assert_called_once_with("cluster_stopped")
        assert resp == EXPECTED_RESPONSE

    @mock.patch.object(NeptuneHook, "conn")
    @mock.patch.object(NeptuneHook, "get_cluster_status")
    @mock.patch.object(NeptuneHook, "get_waiter")
    def test_stop_cluster_no_wait(self, mock_hook_get_waiter, mock_get_cluster_status, mock_conn):
        mock_get_cluster_status.return_value = "available"

        operator = NeptuneStopDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=False,
            wait_for_completion=False,
            aws_conn_id="aws_default",
        )

        resp = operator.execute(None)
        mock_hook_get_waiter.assert_not_called()
        assert resp == EXPECTED_RESPONSE

    @mock.patch.object(NeptuneHook, "conn")
    @mock.patch.object(NeptuneHook, "get_cluster_status")
    @mock.patch.object(NeptuneHook, "get_waiter")
    def test_stop_cluster_cluster_stopped(self, mock_waiter, mock_get_cluster_status, mock_conn):
        mock_get_cluster_status.return_value = "stopped"
        operator = NeptuneStopDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=False,
            wait_for_completion=True,
            aws_conn_id="aws_default",
        )

        resp = operator.execute(None)

        mock_conn.stop_db_cluster.assert_not_called()
        mock_waiter.assert_not_called()
        assert resp == {"db_cluster_id": CLUSTER_ID}

    @mock.patch.object(NeptuneHook, "conn")
    @mock.patch.object(NeptuneHook, "get_cluster_status")
    @mock.patch.object(NeptuneHook, "get_waiter")
    def test_stop_cluster_cluster_error(self, mock_waiter, mock_get_cluster_status, mock_conn):
        mock_get_cluster_status.return_value = "migration-failed"
        operator = NeptuneStopDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=False,
            wait_for_completion=True,
            aws_conn_id="aws_default",
        )

        with pytest.raises(AirflowException):
            operator.execute(None)

    @mock.patch.object(NeptuneHook, "conn")
    @mock.patch.object(NeptuneHook, "get_cluster_status")
    @mock.patch.object(NeptuneHook, "get_waiter")
    def test_stop_cluster_not_in_available(self, mock_waiter, mock_get_cluster_status, mock_conn):
        mock_get_cluster_status.return_value = "backing-up"
        operator = NeptuneStopDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=False,
            wait_for_completion=True,
            aws_conn_id="aws_default",
        )

        operator.execute(None)
        mock_waiter.assert_called_with("cluster_stopped")

    @mock.patch.object(NeptuneStopDbClusterOperator, "defer")
    @mock.patch.object(NeptuneHook, "conn")
    def test_stop_cluster_not_ready_defer(self, mock_conn, mock_defer):
        err_response = {"Error": {"Code": "InvalidClusterState", "Message": "Test message"}}
        exception = client("neptune").exceptions.ClientError(err_response, "test")
        returned_exception = type(exception)

        mock_conn.exceptions.InvalidClusterStateFault = returned_exception
        mock_conn.stop_db_cluster.side_effect = exception
        operator = NeptuneStopDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=True,
            wait_for_completion=False,
            aws_conn_id="aws_default",
        )

        operator.execute(None)
        assert mock_defer.call_count == 2

    @mock.patch.object(NeptuneHook, "conn")
    @mock.patch.object(NeptuneHook, "get_cluster_status")
    @mock.patch.object(NeptuneHook, "get_waiter")
    def test_stop_cluster_instances_not_ready(self, mock_get_waiter, mock_get_cluster_status, mock_conn):
        err_response = {"Error": {"Code": "InvalidDBInstanceState", "Message": "Test message"}}
        exception = client("neptune").exceptions.ClientError(err_response, "test")
        returned_exception = type(exception)

        mock_conn.exceptions.InvalidClusterStateFault = returned_exception
        mock_conn.stop_db_cluster.side_effect = exception

        operator = NeptuneStopDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=False,
            wait_for_completion=False,
            aws_conn_id="aws_default",
        )
        operator.execute(None)
        mock_get_waiter.assert_any_call("db_instance_available")

    @mock.patch("airflow.providers.amazon.aws.operators.neptune.NeptuneStopDbClusterOperator.defer")
    @mock.patch.object(NeptuneHook, "conn")
    def test_stop_cluster_instances_not_ready_defer(self, mock_conn, mock_defer):
        """Tests both waiters are called if an instance exception is raised"""

        err_response = {"Error": {"Code": "InvalidDBInstanceState", "Message": "Test message"}}
        exception = client("neptune").exceptions.ClientError(err_response, "test")
        returned_exception = type(exception)

        mock_conn.exceptions.InvalidClusterStateFault = returned_exception
        mock_conn.stop_db_cluster.side_effect = exception
        operator = NeptuneStopDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=True,
            wait_for_completion=False,
            aws_conn_id="aws_default",
        )

        operator.execute(None)
        # setting the trigger to ANY because an actual trigger doesn't seem to pass the test, even though
        # the output (strings) are exactly the same. Suspect it may be an issue in the trigger's __eq__ function

        assert mock_defer.call_count == 2

    @mock.patch.object(NeptuneHook, "conn")
    def test_stop_cluster_deferrable(self, mock_conn):
        operator = NeptuneStopDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=True,
            wait_for_completion=False,
            aws_conn_id="aws_default",
        )

        with pytest.raises(TaskDeferred):
            operator.execute(None)

    def test_template_fields(self):
        operator = NeptuneStopDbClusterOperator(
            task_id="task_test",
            db_cluster_id=CLUSTER_ID,
            deferrable=True,
            wait_for_completion=False,
            aws_conn_id="aws_default",
        )
        validate_template_fields(operator)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING
from unittest import mock

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID

from airflow.providers.amazon.aws.hooks.glue_crawler import GlueCrawlerHook

if TYPE_CHECKING:
    from unittest.mock import MagicMock

mock_crawler_name = "test-crawler"
mock_role_name = "test-role"
mock_config = {
    "Name": mock_crawler_name,
    "Description": "Test glue crawler from Airflow",
    "DatabaseName": "test_db",
    "Role": mock_role_name,
    "Targets": {
        "S3Targets": [
            {
                "Path": "s3://test-glue-crawler/foo/",
                "Exclusions": [
                    "s3://test-glue-crawler/bar/",
                ],
                "ConnectionName": "test-s3-conn",
            }
        ],
        "JdbcTargets": [
            {
                "ConnectionName": "test-jdbc-conn",
                "Path": "test_db/test_table>",
                "Exclusions": [
                    "string",
                ],
            }
        ],
        "MongoDBTargets": [
            {"ConnectionName": "test-mongo-conn", "Path": "test_db/test_collection", "ScanAll": True}
        ],
        "DynamoDBTargets": [{"Path": "test_db/test_table", "scanAll": True, "scanRate": 123.0}],
        "CatalogTargets": [
            {
                "DatabaseName": "test_glue_db",
                "Tables": [
                    "test",
                ],
            }
        ],
    },
    "Classifiers": ["test-classifier"],
    "TablePrefix": "test",
    "SchemaChangePolicy": {
        "UpdateBehavior": "UPDATE_IN_DATABASE",
        "DeleteBehavior": "DEPRECATE_IN_DATABASE",
    },
    "RecrawlPolicy": {"RecrawlBehavior": "CRAWL_EVERYTHING"},
    "LineageConfiguration": "ENABLE",
    "Configuration": """
    {
        "Version": 1.0,
        "CrawlerOutput": {
            "Partitions": { "AddOrUpdateBehavior": "InheritFromTable" }
        }
    }
    """,
    "SecurityConfiguration": "test",
    "Tags": {"test": "foo", "bar": "test"},
}


class TestGlueCrawlerHook:
    def setup_method(self):
        self.hook = GlueCrawlerHook(aws_conn_id="aws_default")
        self.crawler_arn = (
            f"arn:aws:glue:{self.hook.conn_region_name}:{DEFAULT_ACCOUNT_ID}:crawler/{mock_crawler_name}"
        )

    def test_init(self):
        assert self.hook.aws_conn_id == "aws_default"

    @mock.patch.object(GlueCrawlerHook, "get_conn")
    def test_has_crawler(self, mock_get_conn):
        assert self.hook.has_crawler(mock_crawler_name) is True
        mock_get_conn.return_value.get_crawler.assert_called_once_with(Name=mock_crawler_name)

    @mock.patch.object(GlueCrawlerHook, "get_conn")
    def test_has_crawler_crawled_doesnt_exists(self, mock_get_conn):
        class MockException(Exception):
            pass

        mock_get_conn.return_value.exceptions.EntityNotFoundException = MockException
        mock_get_conn.return_value.get_crawler.side_effect = MockException("AAA")
        assert self.hook.has_crawler(mock_crawler_name) is False
        mock_get_conn.return_value.get_crawler.assert_called_once_with(Name=mock_crawler_name)

    @mock_aws
    @mock.patch.object(GlueCrawlerHook, "get_conn")
    def test_update_crawler_needed(self, mock_get_conn):
        mock_get_conn.return_value.get_crawler.return_value = {"Crawler": mock_config}

        mock_config_two = deepcopy(mock_config)
        mock_config_two["Role"] = "test-2-role"
        mock_config_two.pop("Tags")
        assert self.hook.update_crawler(**mock_config_two) is True
        mock_get_conn.return_value.get_crawler.assert_called_once_with(Name=mock_crawler_name)
        mock_get_conn.return_value.update_crawler.assert_called_once_with(**mock_config_two)

    @mock_aws
    @mock.patch.object(GlueCrawlerHook, "get_conn")
    def test_update_crawler_missing_keys(self, mock_get_conn):
        mock_config_missing_configuration = deepcopy(mock_config)
        mock_config_missing_configuration.pop("Configuration")
        mock_get_conn.return_value.get_crawler.return_value = {"Crawler": mock_config_missing_configuration}

        mock_config_two = deepcopy(mock_config)
        mock_config_two.pop("Tags")
        assert self.hook.update_crawler(**mock_config_two) is True
        mock_get_conn.return_value.get_crawler.assert_called_once_with(Name=mock_crawler_name)
        mock_get_conn.return_value.update_crawler.assert_called_once_with(**mock_config_two)

    @mock_aws
    @mock.patch.object(GlueCrawlerHook, "get_conn")
    def test_update_tags_not_needed(self, mock_get_conn):
        mock_get_conn.return_value.get_crawler.return_value = {"Crawler": mock_config}
        mock_get_conn.return_value.get_tags.return_value = {"Tags": mock_config["Tags"]}

        assert self.hook.update_tags(mock_crawler_name, mock_config["Tags"]) is False
        mock_get_conn.return_value.get_tags.assert_called_once_with(ResourceArn=self.crawler_arn)
        mock_get_conn.return_value.tag_resource.assert_not_called()
        mock_get_conn.return_value.untag_resource.assert_not_called()

    @mock_aws
    @mock.patch.object(GlueCrawlerHook, "get_conn")
    def test_remove_all_tags(self, mock_get_conn):
        mock_get_conn.return_value.get_crawler.return_value = {"Crawler": mock_config}
        mock_get_conn.return_value.get_tags.return_value = {"Tags": mock_config["Tags"]}

        assert self.hook.update_tags(mock_crawler_name, {}) is True
        mock_get_conn.return_value.get_tags.assert_called_once_with(ResourceArn=self.crawler_arn)
        mock_get_conn.return_value.tag_resource.assert_not_called()
        mock_get_conn.return_value.untag_resource.assert_called_once_with(
            ResourceArn=self.crawler_arn, TagsToRemove=["test", "bar"]
        )

    @mock_aws
    @mock.patch.object(GlueCrawlerHook, "get_conn")
    def test_update_missing_tags(self, mock_get_conn):
        mock_config_missing_tags = deepcopy(mock_config)
        mock_config_missing_tags.pop("Tags")
        mock_get_conn.return_value.get_crawler.return_value = {"Crawler": mock_config_missing_tags}

        assert self.hook.update_crawler(**mock_config_missing_tags) is False
        mock_get_conn.return_value.get_tags.assert_not_called()
        mock_get_conn.return_value.tag_resource.assert_not_called()
        mock_get_conn.return_value.untag_resource.assert_not_called()

    @mock_aws
    @mock.patch.object(GlueCrawlerHook, "get_conn")
    def test_replace_tag(self, mock_get_conn):
        mock_get_conn.return_value.get_crawler.return_value = {"Crawler": mock_config}
        mock_get_conn.return_value.get_tags.return_value = {"Tags": mock_config["Tags"]}

        assert self.hook.update_tags(mock_crawler_name, {"test": "bla", "bar": "test"}) is True
        mock_get_conn.return_value.get_tags.assert_called_once_with(ResourceArn=self.crawler_arn)
        mock_get_conn.return_value.untag_resource.assert_not_called()
        mock_get_conn.return_value.tag_resource.assert_called_once_with(
            ResourceArn=self.crawler_arn, TagsToAdd={"test": "bla"}
        )

    @mock_aws
    @mock.patch.object(GlueCrawlerHook, "get_conn")
    def test_update_crawler_not_needed(self, mock_get_conn):
        mock_get_conn.return_value.get_crawler.return_value = {"Crawler": mock_config}
        assert self.hook.update_crawler(**mock_config) is False
        mock_get_conn.return_value.get_crawler.assert_called_once_with(Name=mock_crawler_name)

    @mock.patch.object(GlueCrawlerHook, "get_conn")
    def test_create_crawler(self, mock_get_conn):
        mock_get_conn.return_value.create_crawler.return_value = {"Crawler": {"Name": mock_crawler_name}}
        glue_crawler = self.hook.create_crawler(**mock_config)
        assert "Crawler" in glue_crawler
        assert "Name" in glue_crawler["Crawler"]
        assert glue_crawler["Crawler"]["Name"] == mock_crawler_name

    @mock.patch.object(GlueCrawlerHook, "get_conn")
    def test_start_crawler(self, mock_get_conn):
        result = self.hook.start_crawler(mock_crawler_name)
        assert result == mock_get_conn.return_value.start_crawler.return_value

        mock_get_conn.return_value.start_crawler.assert_called_once_with(Name=mock_crawler_name)

    @mock.patch.object(GlueCrawlerHook, "get_crawler")
    @mock.patch.object(GlueCrawlerHook, "get_conn")
    @mock.patch.object(GlueCrawlerHook, "get_waiter")
    def test_wait_for_crawler_completion_instant_ready(
        self, _, mock_get_conn: MagicMock, mock_get_crawler: MagicMock
    ):
        mock_get_crawler.return_value = {"State": "READY", "LastCrawl": {"Status": "MOCK_STATUS"}}
        mock_get_conn.return_value.get_crawler_metrics.return_value = {
            "CrawlerMetricsList": [
                {
                    "LastRuntimeSeconds": "TEST-A",
                    "MedianRuntimeSeconds": "TEST-B",
                    "TablesCreated": "TEST-C",
                    "TablesUpdated": "TEST-D",
                    "TablesDeleted": "TEST-E",
                }
            ]
        }
        assert self.hook.wait_for_crawler_completion(mock_crawler_name) == "MOCK_STATUS"
        mock_get_conn.assert_has_calls(
            [
                mock.call(),
                mock.call().get_crawler_metrics(CrawlerNameList=[mock_crawler_name]),
            ]
        )
        mock_get_crawler.assert_called_once_with(mock_crawler_name)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.amazon.aws.triggers.glue_databrew import GlueDataBrewJobCompleteTrigger

TEST_JOB_NAME = "test_job_name"
TEST_JOB_RUN_ID = "a1234"
TEST_JOB_RUN_STATUS = "SUCCEEDED"


@pytest.fixture
def trigger():
    return GlueDataBrewJobCompleteTrigger(
        aws_conn_id="aws_default", job_name=TEST_JOB_NAME, run_id=TEST_JOB_RUN_ID
    )


class TestGlueDataBrewJobCompleteTrigger:
    def test_serialize(self, trigger):
        class_path, args = trigger.serialize()

        class_name = class_path.split(".")[-1]
        clazz = globals()[class_name]
        instance = clazz(**args)

        class_path2, args2 = instance.serialize()

        assert class_path == class_path2
        assert args == args2



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.amazon.aws.operators.sns import SnsPublishOperator

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

TASK_ID = "sns_publish_job"
AWS_CONN_ID = "custom_aws_conn"
TARGET_ARN = "test-topic.fifo"
MESSAGE = "Message to send"
SUBJECT = "Subject to send"
MESSAGE_ATTRIBUTES = {"test-attribute": "Attribute to send"}


class TestSnsPublishOperator:
    @pytest.fixture(autouse=True)
    def _setup_test_cases(self):
        self.default_op_kwargs = {
            "task_id": TASK_ID,
            "target_arn": TARGET_ARN,
            "message": MESSAGE,
            "subject": SUBJECT,
            "message_attributes": MESSAGE_ATTRIBUTES,
        }

    def test_init(self):
        op = SnsPublishOperator(**self.default_op_kwargs)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

        op = SnsPublishOperator(
            **self.default_op_kwargs,
            aws_conn_id=AWS_CONN_ID,
            region_name="us-west-1",
            verify="/spam/egg.pem",
            botocore_config={"read_timeout": 42},
            message_deduplication_id="abc",
            message_group_id="a",
        )
        assert op.hook.aws_conn_id == AWS_CONN_ID
        assert op.hook._region_name == "us-west-1"
        assert op.hook._verify == "/spam/egg.pem"
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

    @mock.patch.object(SnsPublishOperator, "hook")
    @pytest.mark.parametrize(
        "message_deduplication_id_,message_group_id_",
        [
            ("abc", "a"),
            (None, None),
            ("abc", None),
            (None, "a"),
        ],
    )
    def test_execute(self, mocked_hook, message_deduplication_id_, message_group_id_):
        hook_response = {"MessageId": "foobar"}
        mocked_hook.publish_to_target.return_value = hook_response

        op = SnsPublishOperator(
            **self.default_op_kwargs,
            message_deduplication_id=message_deduplication_id_,
            message_group_id=message_group_id_,
        )
        assert op.execute({}) == hook_response

        mocked_hook.publish_to_target.assert_called_once_with(
            message=MESSAGE,
            message_attributes=MESSAGE_ATTRIBUTES,
            subject=SUBJECT,
            target_arn=TARGET_ARN,
            message_deduplication_id=message_deduplication_id_,
            message_group_id=message_group_id_,
        )

    def test_template_fields(self):
        operator = SnsPublishOperator(
            **self.default_op_kwargs, message_deduplication_id="abc", message_group_id="a"
        )
        validate_template_fields(operator)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest
from moto import mock_aws

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.ec2 import EC2Hook


class TestEC2Hook:
    def test_init(self):
        ec2_hook = EC2Hook(
            aws_conn_id="aws_conn_test",
            region_name="region-test",
        )
        # We're mocking all actual AWS calls and don't need a connection. This
        # avoids an Airflow warning about connection cannot be found.
        ec2_hook.get_connection = lambda _: None
        assert ec2_hook.aws_conn_id == "aws_conn_test"
        assert ec2_hook.region_name == "region-test"

    @classmethod
    def _create_instances(cls, hook: EC2Hook, max_count=1, min_count=1):
        """Create Instances and return all instance ids."""
        conn = hook.get_conn()
        try:
            ec2_client = conn.meta.client
        except AttributeError:
            ec2_client = conn

        # We need existed AMI Image ID otherwise `moto` will raise DeprecationWarning.
        images = ec2_client.describe_images()["Images"]
        response = ec2_client.run_instances(
            MaxCount=max_count, MinCount=min_count, ImageId=images[0]["ImageId"]
        )
        return [instance["InstanceId"] for instance in response["Instances"]]

    @classmethod
    def _create_instance(cls, hook: EC2Hook):
        """Create Instance and return instance id."""
        return cls._create_instances(hook)[0]

    @mock_aws
    def test_get_conn_returns_boto3_resource(self):
        ec2_hook = EC2Hook()
        instances = list(ec2_hook.conn.instances.all())
        assert instances is not None

    @mock_aws
    def test_client_type_get_conn_returns_boto3_resource(self):
        ec2_hook = EC2Hook(api_type="client_type")
        instances = list(ec2_hook.get_instances())
        assert instances is not None

    @mock_aws
    def test_get_instance(self):
        ec2_hook = EC2Hook()
        created_instance_id = self._create_instance(ec2_hook)
        # test get_instance method
        existing_instance = ec2_hook.get_instance(instance_id=created_instance_id)
        assert created_instance_id == existing_instance.instance_id

    @mock_aws
    def test_get_instance_client_type(self):
        ec2_hook = EC2Hook(api_type="client_type")
        created_instance_id = self._create_instance(ec2_hook)
        # test get_instance method
        existing_instance = ec2_hook.get_instance(instance_id=created_instance_id)
        assert created_instance_id == existing_instance["InstanceId"]

    @mock_aws
    def test_get_instance_state(self):
        ec2_hook = EC2Hook()
        created_instance_id = self._create_instance(ec2_hook)
        all_instances = list(ec2_hook.conn.instances.all())
        created_instance_state = all_instances[0].state["Name"]
        # test get_instance_state method
        existing_instance_state = ec2_hook.get_instance_state(instance_id=created_instance_id)
        assert created_instance_state == existing_instance_state

    @mock_aws
    def test_client_type_get_instance_state(self):
        ec2_hook = EC2Hook(api_type="client_type")
        created_instance_id = self._create_instance(ec2_hook)
        all_instances = ec2_hook.get_instances()
        created_instance_state = all_instances[0]["State"]["Name"]

        existing_instance_state = ec2_hook.get_instance_state(instance_id=created_instance_id)
        assert created_instance_state == existing_instance_state

    @mock_aws
    def test_client_type_start_instances(self):
        ec2_hook = EC2Hook(api_type="client_type")
        created_instance_id = self._create_instance(ec2_hook)
        response = ec2_hook.start_instances(instance_ids=[created_instance_id])

        assert response["StartingInstances"][0]["InstanceId"] == created_instance_id
        assert ec2_hook.get_instance_state(created_instance_id) == "running"

    @mock_aws
    def test_client_type_stop_instances(self):
        ec2_hook = EC2Hook(api_type="client_type")
        created_instance_id = self._create_instance(ec2_hook)
        response = ec2_hook.stop_instances(instance_ids=[created_instance_id])

        assert response["StoppingInstances"][0]["InstanceId"] == created_instance_id
        assert ec2_hook.get_instance_state(created_instance_id) == "stopped"

    @mock_aws
    def test_client_type_terminate_instances(self):
        ec2_hook = EC2Hook(api_type="client_type")
        created_instance_id = self._create_instance(ec2_hook)
        response = ec2_hook.terminate_instances(instance_ids=[created_instance_id])

        assert response["TerminatingInstances"][0]["InstanceId"] == created_instance_id
        assert ec2_hook.get_instance_state(created_instance_id) == "terminated"

    @mock_aws
    def test_client_type_describe_instances(self):
        ec2_hook = EC2Hook(api_type="client_type")
        created_instance_id = self._create_instance(ec2_hook)

        # Without filter
        response = ec2_hook.describe_instances(instance_ids=[created_instance_id])

        assert response["Reservations"][0]["Instances"][0]["InstanceId"] == created_instance_id
        assert response["Reservations"][0]["Instances"][0]["State"]["Name"] == "running"

        # With valid filter
        response = ec2_hook.describe_instances(
            filters=[{"Name": "instance-id", "Values": [created_instance_id]}]
        )

        assert len(response["Reservations"]) == 1
        assert response["Reservations"][0]["Instances"][0]["InstanceId"] == created_instance_id
        assert response["Reservations"][0]["Instances"][0]["State"]["Name"] == "running"

        # With invalid filter
        response = ec2_hook.describe_instances(
            filters=[{"Name": "instance-id", "Values": ["invalid_instance_id"]}]
        )

        assert len(response["Reservations"]) == 0

    @mock_aws
    def test_client_type_get_instances(self):
        ec2_hook = EC2Hook(api_type="client_type")
        created_instances = self._create_instances(ec2_hook, max_count=2, min_count=2)
        created_instance_id_1, created_instance_id_2 = created_instances

        # Without filter
        response = ec2_hook.get_instances(instance_ids=[created_instance_id_1, created_instance_id_2])

        assert response[0]["InstanceId"] == created_instance_id_1
        assert response[1]["InstanceId"] == created_instance_id_2

        # With valid filter
        response = ec2_hook.get_instances(
            filters=[{"Name": "instance-id", "Values": [created_instance_id_1, created_instance_id_2]}]
        )

        assert len(response) == 2
        assert response[0]["InstanceId"] == created_instance_id_1
        assert response[1]["InstanceId"] == created_instance_id_2

        # With filter and instance ids
        response = ec2_hook.get_instances(
            filters=[{"Name": "instance-id", "Values": [created_instance_id_1]}],
            instance_ids=[created_instance_id_1, created_instance_id_2],
        )

        assert len(response) == 1
        assert response[0]["InstanceId"] == created_instance_id_1

        # With invalid filter
        response = ec2_hook.get_instances(
            filters=[{"Name": "instance-id", "Values": ["invalid_instance_id"]}]
        )

        assert len(response) == 0

    @mock_aws
    def test_client_type_get_instance_ids(self):
        ec2_hook = EC2Hook(api_type="client_type")
        created_instances = self._create_instances(ec2_hook, max_count=2, min_count=2)
        created_instance_id_1, created_instance_id_2 = created_instances

        # Without filter
        response = ec2_hook.get_instance_ids()

        assert len(response) == 2
        assert response[0] == created_instance_id_1
        assert response[1] == created_instance_id_2

        # With valid filter
        response = ec2_hook.get_instance_ids(filters=[{"Name": "instance-type", "Values": ["m1.small"]}])

        assert len(response) == 2
        assert response[0] == created_instance_id_1
        assert response[1] == created_instance_id_2

        # With invalid filter
        response = ec2_hook.get_instance_ids(
            filters=[{"Name": "instance-type", "Values": ["invalid_instance_type"]}]
        )

        assert len(response) == 0

    @mock_aws
    def test_decorator_only_client_type(self):
        ec2_hook = EC2Hook()

        # Try calling a method which is only supported by client_type API
        with pytest.raises(AirflowException):
            ec2_hook.get_instances()

        # Explicitly provide resource_type as api_type
        ec2_hook = EC2Hook(api_type="resource_type")

        # Try calling a method which is only supported by client_type API
        with pytest.raises(AirflowException):
            ec2_hook.describe_instances()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import re
from copy import deepcopy
from re import Pattern
from typing import TYPE_CHECKING

from unit.amazon.aws.utils.eks_test_constants import (
    STATUS,
    ClusterAttributes,
    ClusterInputs,
    FargateProfileAttributes,
    FargateProfileInputs,
    NodegroupAttributes,
    NodegroupInputs,
    ResponseAttributes,
)

if TYPE_CHECKING:
    import datetime

    from airflow.providers.amazon.aws.hooks.eks import EksHook

InputTypes = type[ClusterInputs] | type[NodegroupInputs] | type[FargateProfileInputs]


def attributes_to_test(
    inputs: InputTypes,
    cluster_name: str,
    fargate_profile_name: str | None = None,
    nodegroup_name: str | None = None,
) -> list[tuple]:
    """
    Assembles the list of tuples which will be used to validate test results.
    The format of the tuple is (attribute name, expected value)

    :param inputs: A class containing lists of tuples to use for verifying the output
    of cluster or nodegroup creation tests.
    :param cluster_name: The name of the cluster under test.
    :param fargate_profile_name: The name of the Fargate profile under test if applicable.
    :param nodegroup_name: The name of the nodegroup under test if applicable.
    :return: Returns a list of tuples containing the keys and values to be validated in testing.
    """
    result: list[tuple] = deepcopy(inputs.REQUIRED + inputs.OPTIONAL + [STATUS])  # type: ignore
    if inputs == ClusterInputs:
        result += [(ClusterAttributes.NAME, cluster_name)]
    elif inputs == FargateProfileInputs:
        result += [(FargateProfileAttributes.FARGATE_PROFILE_NAME, fargate_profile_name)]
    elif inputs == NodegroupInputs:
        # The below tag is mandatory and must have a value of either 'owned' or 'shared'
        # A value of 'owned' denotes that the subnets are exclusive to the nodegroup.
        # The 'shared' value allows more than one resource to use the subnet.
        required_tag: dict = {f"kubernetes.io/cluster/{cluster_name}": "owned"}
        # Find the user-submitted tag set and append the required tag to it.
        final_tag_set: dict = required_tag
        for key, value in result:
            if key == "tags":
                final_tag_set = {**value, **final_tag_set}
        # Inject it back into the list.
        result = [
            (key, value) if (key != NodegroupAttributes.TAGS) else (NodegroupAttributes.TAGS, final_tag_set)
            for key, value in result
        ]

        result += [(NodegroupAttributes.NODEGROUP_NAME, nodegroup_name)]

    return result


def generate_clusters(eks_hook: EksHook, num_clusters: int, minimal: bool) -> list[str]:
    """
    Generates a number of EKS Clusters with data and adds them to the mocked backend.

    :param eks_hook: An EksHook object used to call the EKS API.
    :param num_clusters: Number of clusters to generate.
    :param minimal: If True, only the required values are generated; if False all values are generated.
    :return: Returns a list of the names of the generated clusters.
    """
    # Generates N clusters named cluster0, cluster1, .., clusterN
    return [
        eks_hook.create_cluster(name=f"cluster{count}", **_input_builder(ClusterInputs, minimal))[
            ResponseAttributes.CLUSTER
        ][ClusterAttributes.NAME]
        for count in range(num_clusters)
    ]


def generate_fargate_profiles(
    eks_hook: EksHook, cluster_name: str, num_profiles: int, minimal: bool
) -> list[str]:
    """
    Generates a number of EKS Fargate profiles with data and adds them to the mocked backend.

    :param eks_hook: An EksHook object used to call the EKS API.
    :param cluster_name: The name of the EKS Cluster to attach the nodegroups to.
    :param num_profiles: Number of Fargate profiles to generate.
    :param minimal: If True, only the required values are generated; if False all values are generated.
    :return: Returns a list of the names of the generated nodegroups.
    """
    # Generates N Fargate profiles named profile0, profile1, .., profileN
    return [
        eks_hook.create_fargate_profile(
            fargateProfileName=f"profile{count}",
            clusterName=cluster_name,
            **_input_builder(FargateProfileInputs, minimal),
        )[ResponseAttributes.FARGATE_PROFILE][FargateProfileAttributes.FARGATE_PROFILE_NAME]
        for count in range(num_profiles)
    ]


def generate_nodegroups(
    eks_hook: EksHook, cluster_name: str, num_nodegroups: int, minimal: bool
) -> list[str]:
    """
    Generates a number of EKS Managed Nodegroups with data and adds them to the mocked backend.

    :param eks_hook: An EksHook object used to call the EKS API.
    :param cluster_name: The name of the EKS Cluster to attach the nodegroups to.
    :param num_nodegroups: Number of clusters to generate.
    :param minimal: If True, only the required values are generated; if False all values are generated.
    :return: Returns a list of the names of the generated nodegroups.
    """
    # Generates N nodegroups named nodegroup0, nodegroup1, .., nodegroupN
    return [
        eks_hook.create_nodegroup(
            nodegroupName=f"nodegroup{count}",
            clusterName=cluster_name,
            **_input_builder(NodegroupInputs, minimal),
        )[ResponseAttributes.NODEGROUP][NodegroupAttributes.NODEGROUP_NAME]
        for count in range(num_nodegroups)
    ]


def region_matches_partition(region: str, partition: str) -> bool:
    """
    Returns True if the provided region and partition are a valid pair.

    :param region: AWS region code to test.
    :param partition: AWS partition code to test.
    :return: Returns True if the provided region and partition are a valid pair.
    """
    valid_matches: list[tuple[str, str]] = [
        ("cn-", "aws-cn"),
        ("us-gov-", "aws-us-gov"),
        ("us-gov-iso-", "aws-iso"),
        ("us-gov-iso-b-", "aws-iso-b"),
    ]

    for prefix, expected_partition in valid_matches:
        if region.startswith(prefix):
            return partition == expected_partition
    return partition == "aws"


def _input_builder(options: InputTypes, minimal: bool) -> dict:
    """
    Assembles the inputs which will be used to generate test object into a dictionary.

    :param options: A class containing lists of tuples to use for to create
    the cluster or nodegroup used in testing.
    :param minimal: If True, only the required values are generated; if False all values are generated.
    :return: Returns a dict containing the keys and values to be validated in testing.
    """
    values: list[tuple] = deepcopy(options.REQUIRED)  # type: ignore
    if not minimal:
        values.extend(deepcopy(options.OPTIONAL))
    return dict(values)  # type: ignore


def string_to_regex(value: str) -> Pattern[str]:
    """
    Converts a string template into a regex template for pattern matching.

    :param value: The template string to convert.
    :returns: Returns a regex pattern
    """
    return re.compile(re.sub(r"[{](.*?)[}]", r"(?P<\1>.+)", value))


def convert_keys(original: dict) -> dict:
    """
    API Input and Output keys are formatted differently.  The EKS Hooks map
    as closely as possible to the API calls, which use camelCase variable
    names, but the Operators match python conventions and use snake_case.
    This method converts the keys of a dict which are in snake_case (input
    format) to camelCase (output format) while leaving the dict values unchanged.

    :param original: Dict which needs the keys converted.
    :value original: Dict
    """
    if "nodegroup_name" in original.keys():
        conversion_map = {
            "cluster_name": "clusterName",
            "cluster_role_arn": "roleArn",
            "nodegroup_subnets": "subnets",
            "subnets": "subnets",
            "nodegroup_name": "nodegroupName",
            "nodegroup_role_arn": "nodeRole",
        }
    elif "fargate_profile_name" in original.keys():
        conversion_map = {
            "cluster_name": "clusterName",
            "fargate_profile_name": "fargateProfileName",
            "subnets": "subnets",
            # The following are "duplicated" because we used the more verbose/descriptive version
            # in the CreateCluster Operator when creating a cluster alongside a Fargate profile, but
            # the more terse version in the CreateFargateProfile Operator for the sake of convenience.
            "pod_execution_role_arn": "podExecutionRoleArn",
            "fargate_pod_execution_role_arn": "podExecutionRoleArn",
            "selectors": "selectors",
            "fargate_selectors": "selectors",
        }
    else:
        conversion_map = {
            "cluster_name": "name",
            "cluster_role_arn": "roleArn",
            "resources_vpc_config": "resourcesVpcConfig",
        }

    return {conversion_map[k] if k in conversion_map else k: v for (k, v) in deepcopy(original).items()}


def iso_date(input_datetime: datetime.datetime) -> str:
    return f"{input_datetime:%Y-%m-%dT%H:%M:%S}Z"


def generate_dict(prefix, count) -> dict:
    return {f"{prefix}_{_count}": str(_count) for _count in range(count)}



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import base64
from unittest import mock
from unittest.mock import Mock, patch

import pytest

from airflow.exceptions import TaskDeferred
from airflow.providers.amazon.aws.hooks.lambda_function import LambdaHook
from airflow.providers.amazon.aws.operators.lambda_function import (
    LambdaCreateFunctionOperator,
    LambdaInvokeFunctionOperator,
)

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

FUNCTION_NAME = "function_name"
PAYLOADS = [
    pytest.param('{"hello": "airflow"}', id="string-payload"),
    pytest.param(b'{"hello": "airflow"}', id="bytes-payload"),
]
ROLE_ARN = "role_arn"
IMAGE_URI = "image_uri"
LOG_RESPONSE = base64.b64encode(b"FOO\n\nBAR\n\n").decode()
BAD_LOG_RESPONSE = LOG_RESPONSE[:-3]
NO_LOG_RESPONSE_SENTINEL = type("NoLogResponseSentinel", (), {})()


class TestLambdaCreateFunctionOperator:
    def test_init(self):
        op = LambdaCreateFunctionOperator(
            task_id="task_test",
            function_name=FUNCTION_NAME,
            role=ROLE_ARN,
            code={
                "ImageUri": IMAGE_URI,
            },
            aws_conn_id="aws_conn_test",
            region_name="foo-bar-1",
            verify="/spam/egg.pem",
            botocore_config={"baz": "qux"},
        )

        assert op.function_name == FUNCTION_NAME
        assert op.role == ROLE_ARN
        assert op.code == {"ImageUri": IMAGE_URI}

        assert op.aws_conn_id == "aws_conn_test"
        assert op.region_name == "foo-bar-1"
        assert op.verify == "/spam/egg.pem"
        assert op.botocore_config == {"baz": "qux"}

    @mock.patch.object(LambdaHook, "create_lambda")
    @mock.patch.object(LambdaHook, "conn")
    def test_create_lambda_without_wait_for_completion(self, mock_hook_conn, mock_hook_create_lambda):
        operator = LambdaCreateFunctionOperator(
            task_id="task_test",
            function_name=FUNCTION_NAME,
            role=ROLE_ARN,
            code={
                "ImageUri": IMAGE_URI,
            },
        )
        operator.execute(None)

        mock_hook_create_lambda.assert_called_once()
        mock_hook_conn.get_waiter.assert_not_called()

    @mock.patch.object(LambdaHook, "create_lambda")
    @mock.patch.object(LambdaHook, "conn")
    @pytest.mark.parametrize(
        "op_kwargs",
        [
            pytest.param({}, id="no-additional-parameters"),
            pytest.param(
                {"region_name": "eu-west-1", "verify": True, "botocore_config": {}},
                id="additional-parameters",
            ),
        ],
    )
    def test_create_lambda_with_wait_for_completion(self, mock_hook_conn, mock_hook_create_lambda, op_kwargs):
        operator = LambdaCreateFunctionOperator(
            task_id="task_test",
            function_name=FUNCTION_NAME,
            role=ROLE_ARN,
            code={
                "ImageUri": IMAGE_URI,
            },
            wait_for_completion=True,
            aws_conn_id="aws_conn_test",
            **op_kwargs,
        )
        operator.execute(None)

        mock_hook_create_lambda.assert_called_once()
        mock_hook_conn.get_waiter.assert_called_once_with("function_active_v2")

    @mock.patch.object(LambdaHook, "create_lambda")
    def test_create_lambda_deferrable(self, _):
        operator = LambdaCreateFunctionOperator(
            task_id="task_test",
            function_name=FUNCTION_NAME,
            role=ROLE_ARN,
            code={
                "ImageUri": IMAGE_URI,
            },
            deferrable=True,
        )
        with pytest.raises(TaskDeferred):
            operator.execute(None)

    @mock.patch.object(LambdaHook, "create_lambda")
    @mock.patch.object(LambdaHook, "conn")
    @pytest.mark.parametrize(
        "config",
        [
            pytest.param(
                {
                    "architectures": ["arm64"],
                    "logging_config": {"LogFormat": "Text", "LogGroup": "/custom/log-group/"},
                    "snap_start": {"ApplyOn": "PublishedVersions"},
                    "ephemeral_storage": {"Size": 1024},
                },
                id="with-config-argument",
            ),
        ],
    )
    def test_create_lambda_using_config_argument(self, mock_hook_conn, mock_hook_create_lambda, config):
        operator = LambdaCreateFunctionOperator(
            task_id="task_test",
            function_name=FUNCTION_NAME,
            role=ROLE_ARN,
            code={
                "ImageUri": IMAGE_URI,
            },
            config=config,
        )
        operator.execute(None)

        mock_hook_create_lambda.assert_called_once()
        mock_hook_conn.get_waiter.assert_not_called()
        assert operator.config.get("logging_config") == config.get("logging_config")
        assert operator.config.get("architectures") == config.get("architectures")
        assert operator.config.get("snap_start") == config.get("snap_start")
        assert operator.config.get("ephemeral_storage") == config.get("ephemeral_storage")

    def test_template_fields(self):
        operator = LambdaCreateFunctionOperator(
            task_id="task_test",
            function_name=FUNCTION_NAME,
            role=ROLE_ARN,
            code={
                "ImageUri": IMAGE_URI,
            },
        )
        validate_template_fields(operator)


class TestLambdaInvokeFunctionOperator:
    @pytest.mark.parametrize("payload", PAYLOADS)
    def test_init(self, payload):
        lambda_operator = LambdaInvokeFunctionOperator(
            task_id="test",
            function_name="test",
            payload=payload,
            log_type="None",
            aws_conn_id="aws_conn_test",
            region_name="foo-bar-1",
            verify="/spam/egg.pem",
            botocore_config={"baz": "qux"},
        )
        assert lambda_operator.task_id == "test"
        assert lambda_operator.function_name == "test"
        assert lambda_operator.payload == payload
        assert lambda_operator.log_type == "None"
        assert lambda_operator.aws_conn_id == "aws_conn_test"
        assert lambda_operator.region_name == "foo-bar-1"
        assert lambda_operator.verify == "/spam/egg.pem"
        assert lambda_operator.botocore_config == {"baz": "qux"}

    @mock.patch.object(LambdaHook, "invoke_lambda")
    @mock.patch.object(LambdaHook, "conn")
    @pytest.mark.parametrize(
        "keep_empty_log_lines", [pytest.param(True, id="keep"), pytest.param(False, id="truncate")]
    )
    @pytest.mark.parametrize(
        "log_result, expected_execution_logs",
        [
            pytest.param(LOG_RESPONSE, True, id="log-result"),
            pytest.param(BAD_LOG_RESPONSE, False, id="corrupted-log-result"),
            pytest.param(None, False, id="none-log-result"),
            pytest.param(NO_LOG_RESPONSE_SENTINEL, False, id="no-response"),
        ],
    )
    @pytest.mark.parametrize("payload", PAYLOADS)
    def test_invoke_lambda(
        self,
        mock_conn,
        mock_invoke,
        payload,
        keep_empty_log_lines,
        log_result,
        expected_execution_logs,
        caplog,
    ):
        operator = LambdaInvokeFunctionOperator(
            task_id="task_test",
            function_name="a",
            invocation_type="b",
            log_type="c",
            keep_empty_log_lines=keep_empty_log_lines,
            client_context="d",
            payload=payload,
            qualifier="f",
        )
        returned_payload = Mock()
        returned_payload.read().decode.return_value = "data was read"
        fake_response = {
            "ResponseMetadata": "",
            "StatusCode": 200,
            "Payload": returned_payload,
        }
        if log_result is not NO_LOG_RESPONSE_SENTINEL:
            fake_response["LogResult"] = log_result
        mock_invoke.return_value = fake_response

        caplog.set_level("INFO", "airflow.task.operators")
        value = operator.execute(None)

        assert value == "data was read"
        mock_invoke.assert_called_once_with(
            function_name="a",
            invocation_type="b",
            log_type="c",
            client_context="d",
            payload=payload,
            qualifier="f",
        )

        # Validate log messages in task logs
        if expected_execution_logs:
            assert "The last 4 KB of the Lambda execution log" in caplog.text
            assert "FOO" in caplog.messages
            assert "BAR" in caplog.messages
            if keep_empty_log_lines:
                assert "" in caplog.messages
            else:
                assert "" not in caplog.messages
        else:
            assert "The last 4 KB of the Lambda execution log" not in caplog.text

    @patch.object(LambdaInvokeFunctionOperator, "hook", new_callable=mock.PropertyMock)
    def test_invoke_lambda_bad_http_code(self, hook_mock):
        operator = LambdaInvokeFunctionOperator(
            task_id="task_test",
            function_name="a",
        )
        hook_mock().invoke_lambda.return_value = {"ResponseMetadata": "", "StatusCode": 404}

        with pytest.raises(ValueError):
            operator.execute(None)

    @patch.object(LambdaInvokeFunctionOperator, "hook", new_callable=mock.PropertyMock)
    def test_invoke_lambda_function_error(self, hook_mock):
        operator = LambdaInvokeFunctionOperator(
            task_id="task_test",
            function_name="a",
        )
        hook_mock().invoke_lambda.return_value = {
            "ResponseMetadata": "",
            "StatusCode": 404,
            "FunctionError": "yes",
            "Payload": Mock(),
        }

        with pytest.raises(ValueError):
            operator.execute(None)

    def test_template_fields(self):
        operator = LambdaInvokeFunctionOperator(
            task_id="task_test",
            function_name="a",
        )
        validate_template_fields(operator)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import asyncio
from unittest import mock

import pytest

from airflow.providers.amazon.aws.triggers.redshift_cluster import (
    RedshiftClusterTrigger,
)
from airflow.triggers.base import TriggerEvent

POLLING_PERIOD_SECONDS = 1.0


class TestRedshiftClusterTrigger:
    def test_redshift_cluster_sensor_trigger_serialization(self):
        """
        Asserts that the RedshiftClusterTrigger correctly serializes its arguments
        and classpath.
        """
        trigger = RedshiftClusterTrigger(
            aws_conn_id="test_redshift_conn_id",
            cluster_identifier="mock_cluster_identifier",
            target_status="available",
            poke_interval=POLLING_PERIOD_SECONDS,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.redshift_cluster.RedshiftClusterTrigger"
        assert kwargs == {
            "aws_conn_id": "test_redshift_conn_id",
            "cluster_identifier": "mock_cluster_identifier",
            "target_status": "available",
            "poke_interval": POLLING_PERIOD_SECONDS,
        }

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_cluster.RedshiftHook.cluster_status_async")
    async def test_redshift_cluster_sensor_trigger_success(self, mock_cluster_status):
        """
        Test RedshiftClusterTrigger with the success status
        """
        expected_result = "available"

        mock_cluster_status.return_value = expected_result
        trigger = RedshiftClusterTrigger(
            aws_conn_id="test_redshift_conn_id",
            cluster_identifier="mock_cluster_identifier",
            target_status="available",
            poke_interval=POLLING_PERIOD_SECONDS,
        )

        generator = trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "success", "message": "target state met"}) == actual

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "expected_result",
        ["Resuming"],
    )
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_cluster.RedshiftHook.cluster_status_async")
    async def test_redshift_cluster_sensor_trigger_resuming_status(
        self, mock_cluster_status, expected_result
    ):
        """Test RedshiftClusterTrigger with the success status"""
        mock_cluster_status.return_value = expected_result
        trigger = RedshiftClusterTrigger(
            aws_conn_id="test_redshift_conn_id",
            cluster_identifier="mock_cluster_identifier",
            target_status="available",
            poke_interval=POLLING_PERIOD_SECONDS,
        )
        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # TriggerEvent was returned
        assert task.done() is False

        # Prevents error when task is destroyed while in "pending" state
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_cluster.RedshiftHook.cluster_status_async")
    async def test_redshift_cluster_sensor_trigger_exception(self, mock_cluster_status):
        """Test RedshiftClusterTrigger with exception"""
        mock_cluster_status.side_effect = Exception("Test exception")
        trigger = RedshiftClusterTrigger(
            aws_conn_id="test_redshift_conn_id",
            cluster_identifier="mock_cluster_identifier",
            target_status="available",
            poke_interval=POLLING_PERIOD_SECONDS,
        )

        task = [i async for i in trigger.run()]
        # since we use return as soon as we yield the trigger event
        # at any given point there should be one trigger event returned to the task
        # so we validate for length of task to be 1
        assert len(task) == 1
        assert TriggerEvent({"status": "error", "message": "Test exception"}) in task



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from datetime import datetime

import boto3

from airflow.decorators import task
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.operators.appflow import (
    AppflowRunOperator,
)
from airflow.providers.amazon.aws.operators.s3 import (
    S3CreateBucketOperator,
    S3CreateObjectOperator,
    S3DeleteBucketOperator,
)
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import SystemTestContextBuilder

sys_test_context_task = SystemTestContextBuilder().build()

DAG_ID = "example_appflow_run"


@task
def create_s3_to_s3_flow(flow_name: str, bucket_name: str, source_folder: str):
    """creates a flow that takes a CSV and converts it to a json containing the same data"""
    client = boto3.client("appflow")
    client.create_flow(
        flowName=flow_name,
        triggerConfig={"triggerType": "OnDemand"},
        sourceFlowConfig={
            "connectorType": "S3",
            "sourceConnectorProperties": {
                "S3": {
                    "bucketName": bucket_name,
                    "bucketPrefix": source_folder,
                    "s3InputFormatConfig": {"s3InputFileType": "CSV"},
                },
            },
        },
        destinationFlowConfigList=[
            {
                "connectorType": "S3",
                "destinationConnectorProperties": {
                    "S3": {
                        "bucketName": bucket_name,
                        "s3OutputFormatConfig": {
                            "fileType": "JSON",
                            "aggregationConfig": {
                                "aggregationType": "None",
                            },
                        },
                    }
                },
            },
        ],
        tasks=[
            {
                "sourceFields": ["col1", "col2"],
                "connectorOperator": {"S3": "PROJECTION"},
                "taskType": "Filter",
            },
            {
                "sourceFields": ["col1"],
                "connectorOperator": {"S3": "NO_OP"},
                "destinationField": "col1",
                "taskType": "Map",
                "taskProperties": {"DESTINATION_DATA_TYPE": "string", "SOURCE_DATA_TYPE": "string"},
            },
            {
                "sourceFields": ["col2"],
                "connectorOperator": {"S3": "NO_OP"},
                "destinationField": "col2",
                "taskType": "Map",
                "taskProperties": {"DESTINATION_DATA_TYPE": "string", "SOURCE_DATA_TYPE": "string"},
            },
        ],
    )


@task
def setup_bucket_permissions(bucket_name):
    s3 = boto3.client("s3")
    s3.put_bucket_policy(
        Bucket=bucket_name,
        Policy=json.dumps(
            {
                "Version": "2008-10-17",
                "Statement": [
                    {
                        "Sid": "AllowAppFlowSourceActions",
                        "Effect": "Allow",
                        "Principal": {"Service": "appflow.amazonaws.com"},
                        "Action": ["s3:ListBucket", "s3:GetObject"],
                        "Resource": [f"arn:aws:s3:::{bucket_name}", f"arn:aws:s3:::{bucket_name}/*"],
                    },
                    {
                        "Sid": "AllowAppFlowDestinationActions",
                        "Effect": "Allow",
                        "Principal": {"Service": "appflow.amazonaws.com"},
                        "Action": [
                            "s3:PutObject",
                            "s3:AbortMultipartUpload",
                            "s3:ListMultipartUploadParts",
                            "s3:ListBucketMultipartUploads",
                            "s3:GetBucketAcl",
                            "s3:PutObjectAcl",
                        ],
                        "Resource": [f"arn:aws:s3:::{bucket_name}", f"arn:aws:s3:::{bucket_name}/*"],
                    },
                ],
            }
        ),
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_flow(flow_name: str):
    client = boto3.client("appflow")
    client.delete_flow(flowName=flow_name, forceDelete=True)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2022, 1, 1),
    catchup=False,
    tags=["example"],
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]

    flow_name = f"{env_id}-flow"
    bucket_name = f"{env_id}-for-appflow"
    source_folder = "source"

    create_bucket = S3CreateBucketOperator(task_id="create_bucket", bucket_name=bucket_name)

    upload_csv = S3CreateObjectOperator(
        task_id="upload_csv",
        s3_bucket=bucket_name,
        s3_key="source_folder/data.csv",
        data="""col1,col2\n"data1","data2"\n""",
        replace=True,
    )

    # [START howto_operator_appflow_run]
    run_flow = AppflowRunOperator(
        task_id="run_flow",
        flow_name=flow_name,
    )
    # [END howto_operator_appflow_run]
    run_flow.poll_interval = 1

    delete_bucket = S3DeleteBucketOperator(
        task_id="delete_bucket",
        trigger_rule=TriggerRule.ALL_DONE,
        bucket_name=bucket_name,
        force_delete=True,
    )

    chain(
        # TEST SETUP
        test_context,
        create_bucket,
        setup_bucket_permissions(bucket_name),
        upload_csv,
        create_s3_to_s3_flow(flow_name, bucket_name, source_folder),
        # TEST BODY
        run_flow,
        # TEARDOWN
        delete_flow(flow_name),
        delete_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains AWS CloudFormation create/delete stack operators."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.providers.amazon.aws.hooks.cloud_formation import CloudFormationHook
from airflow.providers.amazon.aws.operators.base_aws import AwsBaseOperator
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields

if TYPE_CHECKING:
    from airflow.utils.context import Context


class CloudFormationCreateStackOperator(AwsBaseOperator[CloudFormationHook]):
    """
    An operator that creates a AWS CloudFormation stack.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudFormationCreateStackOperator`

    :param stack_name: stack name (templated)
    :param cloudformation_parameters: parameters to be passed to AWS CloudFormation.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = CloudFormationHook
    template_fields: Sequence[str] = aws_template_fields("stack_name", "cloudformation_parameters")
    ui_color = "#6b9659"

    def __init__(self, *, stack_name: str, cloudformation_parameters: dict, **kwargs):
        super().__init__(**kwargs)
        self.stack_name = stack_name
        self.cloudformation_parameters = cloudformation_parameters

    def execute(self, context: Context):
        self.log.info("CloudFormation parameters: %s", self.cloudformation_parameters)
        self.hook.create_stack(self.stack_name, self.cloudformation_parameters)


class CloudFormationDeleteStackOperator(AwsBaseOperator[CloudFormationHook]):
    """
    An operator that deletes a AWS CloudFormation stack.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudFormationDeleteStackOperator`

    :param stack_name: stack name (templated)
    :param cloudformation_parameters: parameters to be passed to CloudFormation.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = CloudFormationHook
    template_fields: Sequence[str] = aws_template_fields("stack_name")
    ui_color = "#1d472b"
    ui_fgcolor = "#FFF"

    def __init__(
        self,
        *,
        stack_name: str,
        cloudformation_parameters: dict | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.cloudformation_parameters = cloudformation_parameters or {}
        self.stack_name = stack_name

    def execute(self, context: Context):
        self.log.info("CloudFormation Parameters: %s", self.cloudformation_parameters)
        self.hook.delete_stack(self.stack_name, self.cloudformation_parameters)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest import mock

import pytest

from airflow.models import Connection
from airflow.providers.amazon.aws.hooks.chime import ChimeWebhookHook
from airflow.providers.amazon.aws.notifications.chime import ChimeNotifier, send_chime_notification
from airflow.providers.standard.operators.empty import EmptyOperator

pytestmark = pytest.mark.db_test


class TestChimeNotifier:
    # Chime webhooks can't really have a default connection, so we need to create one for tests.
    @pytest.fixture(autouse=True)
    def setup_connections(self, create_connection_without_db):
        create_connection_without_db(
            Connection(
                conn_id="default-chime-webhook",
                conn_type="chime",
                host="hooks.chime.aws/incomingwebhooks/",
                password="abcd-1134-ZeDA?token=somechimetoken111",
                schema="https",
            )
        )

    @mock.patch.object(ChimeWebhookHook, "send_message")
    def test_chime_notifier(self, mock_chime_hook, dag_maker):
        with dag_maker("test_chime_notifier") as dag:
            EmptyOperator(task_id="task1")

        notifier = send_chime_notification(
            chime_conn_id="default-chime-webhook", message="Chime Test Message"
        )
        notifier({"dag": dag})
        mock_chime_hook.assert_called_once_with(message="Chime Test Message")

    @mock.patch.object(ChimeWebhookHook, "send_message")
    def test_chime_notifier_with_notifier_class(self, mock_chime_hook, dag_maker):
        with dag_maker("test_chime_notifier") as dag:
            EmptyOperator(task_id="task1")

        notifier = ChimeNotifier(
            chime_conn_id="default-chime-webhook", message="Test Chime Message for Class"
        )
        notifier({"dag": dag})
        mock_chime_hook.assert_called_once_with(message="Test Chime Message for Class")

    @mock.patch.object(ChimeWebhookHook, "send_message")
    def test_chime_notifier_templated(self, mock_chime_hook, dag_maker):
        with dag_maker("test_chime_notifier") as dag:
            EmptyOperator(task_id="task1")

        notifier = send_chime_notification(
            chime_conn_id="default-chime-webhook", message="Test Chime Message. Dag is {{ dag.dag_id }}."
        )
        notifier({"dag": dag})
        mock_chime_hook.assert_called_once_with(message="Test Chime Message. Dag is test_chime_notifier.")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.auth_manager.avp.entities import AvpEntities, get_action_id, get_entity_type
from airflow.providers.amazon.aws.auth_manager.avp.facade import AwsAuthManagerAmazonVerifiedPermissionsFacade
from airflow.providers.amazon.aws.auth_manager.user import AwsAuthManagerUser
from airflow.utils.helpers import prune_dict

from tests_common.test_utils.config import conf_vars

if TYPE_CHECKING:
    from airflow.api_fastapi.auth.managers.base_auth_manager import ResourceMethod

REGION_NAME = "us-east-1"
AVP_POLICY_STORE_ID = "store_id"

test_user = AwsAuthManagerUser(user_id="test_user", groups=["group1", "group2"])
test_user_no_group = AwsAuthManagerUser(user_id="test_user_no_group", groups=[])


@pytest.fixture
def facade():
    with conf_vars(
        {
            ("aws_auth_manager", "conn_id"): "aws_default",
            ("aws_auth_manager", "region_name"): REGION_NAME,
            ("aws_auth_manager", "avp_policy_store_id"): AVP_POLICY_STORE_ID,
        }
    ):
        yield AwsAuthManagerAmazonVerifiedPermissionsFacade()


class TestAwsAuthManagerAmazonVerifiedPermissionsFacade:
    def test_avp_client(self, facade):
        assert hasattr(facade, "avp_client")

    def test_avp_policy_store_id(self, facade):
        assert hasattr(facade, "avp_policy_store_id")

    @pytest.mark.parametrize(
        "entity_id, context, user, expected_entities, expected_context, avp_response, expected",
        [
            # User with groups with no permissions
            (
                None,
                None,
                test_user,
                [
                    {
                        "identifier": {"entityType": "Airflow::User", "entityId": "test_user"},
                        "parents": [
                            {"entityType": "Airflow::Group", "entityId": "group1"},
                            {"entityType": "Airflow::Group", "entityId": "group2"},
                        ],
                    },
                    {
                        "identifier": {"entityType": "Airflow::Group", "entityId": "group1"},
                    },
                    {
                        "identifier": {"entityType": "Airflow::Group", "entityId": "group2"},
                    },
                ],
                None,
                {"decision": "DENY"},
                False,
            ),
            # User with groups with permissions
            (
                "dummy_id",
                None,
                test_user,
                [
                    {
                        "identifier": {"entityType": "Airflow::User", "entityId": "test_user"},
                        "parents": [
                            {"entityType": "Airflow::Group", "entityId": "group1"},
                            {"entityType": "Airflow::Group", "entityId": "group2"},
                        ],
                    },
                    {
                        "identifier": {"entityType": "Airflow::Group", "entityId": "group1"},
                    },
                    {
                        "identifier": {"entityType": "Airflow::Group", "entityId": "group2"},
                    },
                ],
                None,
                {"decision": "ALLOW"},
                True,
            ),
            # User without group without permission
            (
                None,
                None,
                test_user_no_group,
                [
                    {
                        "identifier": {"entityType": "Airflow::User", "entityId": "test_user_no_group"},
                        "parents": [],
                    },
                ],
                None,
                {"decision": "DENY"},
                False,
            ),
            # With context
            (
                "dummy_id",
                {"context_param": {"string": "value"}},
                test_user,
                [
                    {
                        "identifier": {"entityType": "Airflow::User", "entityId": "test_user"},
                        "parents": [
                            {"entityType": "Airflow::Group", "entityId": "group1"},
                            {"entityType": "Airflow::Group", "entityId": "group2"},
                        ],
                    },
                    {
                        "identifier": {"entityType": "Airflow::Group", "entityId": "group1"},
                    },
                    {
                        "identifier": {"entityType": "Airflow::Group", "entityId": "group2"},
                    },
                ],
                {"contextMap": {"context_param": {"string": "value"}}},
                {"decision": "ALLOW"},
                True,
            ),
        ],
    )
    def test_is_authorized_successful(
        self, facade, entity_id, context, user, expected_entities, expected_context, avp_response, expected
    ):
        mock_is_authorized = Mock(return_value=avp_response)
        facade.avp_client.is_authorized = mock_is_authorized

        method: ResourceMethod = "GET"
        entity_type = AvpEntities.VARIABLE

        result = facade.is_authorized(
            method=method,
            entity_type=entity_type,
            entity_id=entity_id,
            user=user,
            context=context,
        )

        params = prune_dict(
            {
                "policyStoreId": AVP_POLICY_STORE_ID,
                "principal": {"entityType": "Airflow::User", "entityId": user.get_id()},
                "action": {"actionType": "Airflow::Action", "actionId": get_action_id(entity_type, method)},
                "resource": {"entityType": get_entity_type(entity_type), "entityId": entity_id or "*"},
                "entities": {"entityList": expected_entities},
                "context": expected_context,
            }
        )

        mock_is_authorized.assert_called_once_with(**params)

        assert result == expected

    def test_is_authorized_unsuccessful(self, facade):
        avp_response = {"errors": ["Error"]}
        mock_is_authorized = Mock(return_value=avp_response)
        facade.avp_client.is_authorized = mock_is_authorized

        with pytest.raises(AirflowException, match="Error occurred while making an authorization decision."):
            facade.is_authorized(method="GET", entity_type=AvpEntities.VARIABLE, user=test_user)

    @pytest.mark.parametrize(
        "user, avp_response, expected",
        [
            (
                test_user,
                {"results": [{"decision": "ALLOW"}, {"decision": "DENY"}]},
                False,
            ),
            (
                test_user,
                {"results": [{"decision": "ALLOW"}, {"decision": "ALLOW"}]},
                True,
            ),
            (
                None,
                {"results": [{"decision": "ALLOW"}, {"decision": "ALLOW"}]},
                False,
            ),
        ],
    )
    def test_batch_is_authorized_successful(self, facade, user, avp_response, expected):
        mock_batch_is_authorized = Mock(return_value=avp_response)
        facade.avp_client.batch_is_authorized = mock_batch_is_authorized

        result = facade.batch_is_authorized(
            requests=[
                {"method": "GET", "entity_type": AvpEntities.VARIABLE, "entity_id": "var1"},
                {"method": "GET", "entity_type": AvpEntities.VARIABLE, "entity_id": "var1"},
            ],
            user=user,
        )

        assert result == expected

    def test_batch_is_authorized_unsuccessful(self, facade):
        avp_response = {"results": [{}, {"errors": []}, {"errors": [{"errorDescription": "Error"}]}]}
        mock_batch_is_authorized = Mock(return_value=avp_response)
        facade.avp_client.batch_is_authorized = mock_batch_is_authorized

        with pytest.raises(
            AirflowException, match="Error occurred while making a batch authorization decision."
        ):
            facade.batch_is_authorized(
                requests=[
                    {"method": "GET", "entity_type": AvpEntities.VARIABLE, "entity_id": "var1"},
                    {"method": "GET", "entity_type": AvpEntities.VARIABLE, "entity_id": "var1"},
                ],
                user=test_user,
            )

    def test_get_batch_is_authorized_single_result_successful(self, facade):
        single_result = {
            "request": {
                "principal": {"entityType": "Airflow::User", "entityId": "test_user"},
                "action": {"actionType": "Airflow::Action", "actionId": "Connection.GET"},
                "resource": {"entityType": "Airflow::Connection", "entityId": "*"},
            },
            "decision": "ALLOW",
        }

        result = facade.get_batch_is_authorized_single_result(
            batch_is_authorized_results=[
                {
                    "request": {
                        "principal": {"entityType": "Airflow::User", "entityId": "test_user"},
                        "action": {"actionType": "Airflow::Action", "actionId": "Variable.GET"},
                        "resource": {"entityType": "Airflow::Variable", "entityId": "*"},
                    },
                    "decision": "ALLOW",
                },
                single_result,
            ],
            request={
                "method": "GET",
                "entity_type": AvpEntities.CONNECTION,
            },
            user=test_user,
        )

        assert result == single_result

    def test_get_batch_is_authorized_single_result_unsuccessful(self, facade):
        with pytest.raises(AirflowException, match="Could not find the authorization result."):
            facade.get_batch_is_authorized_single_result(
                batch_is_authorized_results=[
                    {
                        "request": {
                            "principal": {"entityType": "Airflow::User", "entityId": "test_user"},
                            "action": {"actionType": "Airflow::Action", "actionId": "Variable.GET"},
                            "resource": {"entityType": "Airflow::Variable", "entityId": "*"},
                        },
                        "decision": "ALLOW",
                    },
                    {
                        "request": {
                            "principal": {"entityType": "Airflow::User", "entityId": "test_user"},
                            "action": {"actionType": "Airflow::Action", "actionId": "Variable.POST"},
                            "resource": {"entityType": "Airflow::Variable", "entityId": "*"},
                        },
                        "decision": "ALLOW",
                    },
                ],
                request={
                    "method": "GET",
                    "entity_type": AvpEntities.CONNECTION,
                },
                user=test_user,
            )

    def test_is_policy_store_schema_up_to_date_when_schema_up_to_date(self, facade):
        from airflow.providers.amazon.aws.auth_manager import avp

        schema_path = Path(avp.__file__).parent / "schema.json"
        with open(schema_path) as schema_file:
            avp_response = {"schema": schema_file.read()}
            mock_get_schema = Mock(return_value=avp_response)
            facade.avp_client.get_schema = mock_get_schema

            assert facade.is_policy_store_schema_up_to_date()

    def test_is_policy_store_schema_up_to_date_when_schema_is_modified(self, facade):
        from airflow.providers.amazon.aws.auth_manager import avp

        schema_path = Path(avp.__file__).parent / "schema.json"
        with open(schema_path) as schema_file:
            schema = json.loads(schema_file.read())
            schema["new_field"] = "new_value"
            avp_response = {"schema": json.dumps(schema)}
            mock_get_schema = Mock(return_value=avp_response)
            facade.avp_client.get_schema = mock_get_schema

            assert not facade.is_policy_store_schema_up_to_date()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.amazon.aws.links.step_function import (
    StateMachineDetailsLink,
    StateMachineExecutionsDetailsLink,
)

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS
from unit.amazon.aws.links.test_base_aws import BaseAwsLinksTestCase

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk.execution_time.comms import XComResult

pytestmark = pytest.mark.db_test


class TestStateMachineDetailsLink(BaseAwsLinksTestCase):
    link_class = StateMachineDetailsLink

    @pytest.mark.parametrize(
        "state_machine_arn, expected_url",
        [
            pytest.param("", "", id="empty-arn"),
            pytest.param(None, "", id="arn-not-set"),
            pytest.param(
                "foo:bar",
                "https://console.aws.amazon.com/states/home?region=eu-west-1#/statemachines/view/foo%3Abar",
                id="arn-set",
            ),
        ],
    )
    def test_extra_link(self, state_machine_arn, expected_url: str, mock_supervisor_comms):
        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key=self.link_class.key,
                value={
                    "region_name": "eu-west-1",
                    "aws_domain": self.link_class.get_aws_domain("aws"),
                    "aws_partition": "aws",
                    "state_machine_arn": state_machine_arn,
                },
            )
        self.assert_extra_link_url(
            expected_url=expected_url,
            region_name="eu-west-1",
            aws_partition="aws",
            state_machine_arn=state_machine_arn,
        )


class TestStateMachineExecutionsDetailsLink(BaseAwsLinksTestCase):
    link_class = StateMachineExecutionsDetailsLink

    @pytest.mark.parametrize(
        "execution_arn, expected_url",
        [
            pytest.param("", "", id="empty-arn"),
            pytest.param(None, "", id="arn-not-set"),
            pytest.param(
                "spam:egg:000000000",
                "https://console.aws.amazon.com/states/home?region=eu-west-1"
                "#/v2/executions/details/spam%3Aegg%3A000000000",
                id="arn-set",
            ),
        ],
    )
    def test_extra_link(self, execution_arn, expected_url: str, mock_supervisor_comms):
        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key=self.link_class.key,
                value={
                    "region_name": "eu-west-1",
                    "aws_domain": self.link_class.get_aws_domain("aws"),
                    "aws_partition": "aws",
                    "execution_arn": execution_arn,
                },
            )
        self.assert_extra_link_url(
            expected_url=expected_url,
            region_name="eu-west-1",
            aws_partition="aws",
            execution_arn=execution_arn,
        )
